# -*- coding: utf-8 -*-
'''base base knife mixins'''

from operator import truth
from threading import local
from collections import deque

from stuf.six import identity
from stuf.utils import memoize
from stuf.patterns import searcher

SLOTS = (
    '_in _work _hold _out _original _baseline _each _kw _history _worker '
    '_wrapper _args _pipe'
).split()


class _KnifeMixin(local):

    '''Base knife mixin.'''

    def __init__(self, ins, outs, **kw):
        super(_KnifeMixin, self).__init__()
        # incoming things
        self._in = ins
        # outgoing things
        self._out = outs
        # pipe out default
        self._pipe = None
        # default output default
        self._each = False
        # original and baseline snapshots
        self._original = self._baseline = None
        # maximum number of history snapshots to keep (default: 5)
        self._history = deque(maxlen=kw.pop('snapshots', 5))
        # worker default
        self._worker = None
        # position arguments default
        self._args = ()
        # keyword arguments default
        self._kw = {}
        # default wrapper default
        self._wrapper = list

    @property
    def _identity(self):
        # use  generic identity function for worker if no worker assigned
        return self._worker if self._worker is not None else identity

    @property
    def _test(self, truth_=truth):
        # use truth operator function for worker if no worker assigned
        return self._worker if self._worker is not None else truth_

    @staticmethod
    @memoize
    def _pattern(pat, type, flags, s=searcher):
        # compile glob pattern into regex
        return s((type, pat), flags)

    _REPR = '{0}.{1} ([IN: ({2}) => WORK: ({3}) => HOLD: ({4}) => OUT: ({5})])'

    def _clearsp(self):
        # clear fetch snapshots
        self._history.clear()
        return self



# -*- coding: utf-8 -*-
'''specific knife mixins'''

from math import fsum
from copy import deepcopy
from threading import local
from functools import reduce
from inspect import getmro, isclass
from random import randrange, shuffle
from collections import deque, namedtuple
from operator import attrgetter, itemgetter, methodcaller, truediv
from itertools import (
    chain, combinations, groupby, islice, repeat, permutations, starmap, tee)

from stuf.deep import selfname
from stuf.utils import memoize
from stuf.six.moves import filterfalse, zip_longest  # @UnresolvedImport
from stuf.iterable import deferfunc, deferiter, count
from stuf.collects import OrderedDict, Counter, ChainMap
from stuf.six import filter, items, keys, map, isstring, values, next

Count = namedtuple('Count', 'least most overall')
GroupBy = namedtuple('Group', 'keys groups')
MinMax = namedtuple('MinMax', 'min max')
TrueFalse = namedtuple('TrueFalse', 'true false')


def slicer(x, y, iz=islice, nx=next):
    return nx(iz(x, y, None))


class _CmpMixin(local):

    @memoize
    def _all(self, a=all, m=map):
        # invoke worker on each item to yield truth
        return self._append(a(m(self._test, self._iterable)))

    @memoize
    def _any(self, a=any, m=map):
        # invoke worker on each item to yield truth
        return self._append(a(m(self._test, self._iterable)))

    @memoize
    def _difference(self, symmetric, rz=reduce, se=set):
        if symmetric:
            test = lambda x, y: se(x).symmetric_difference(y)
        else:
            test = lambda x, y: se(x).difference(y)
        return self._xtend(rz(test, self._iterable))

    def _intersection(self, rz=reduce, se=set):
        return self._xtend(
            rz(lambda x, y: se(x).intersection(y), self._iterable)
        )

    def _union(self, rz=reduce, se=set):
        return self._xtend(rz(lambda x, y: se(x).union(y), self._iterable))

    @memoize
    def _unique(self, nx=next, se=set, S=StopIteration):
        def unique(key, iterable):
            seen = se()
            seenadd = seen.add
            try:
                while 1:
                    element = key(nx(iterable))
                    if element not in seen:
                        yield element
                        seenadd(element)
            except S:
                pass
        return self._xtend(unique(self._identity, self._iterable))


class _MathMixin(local):

    def _average(self, cnt=count, su=sum, td=truediv, t=tee):
        i1, i2 = t(self._iterable)
        return self._append(td(su(i1, 0.0), cnt(i2)))

    def _count(self, R=Counter, T=Count):
        cnt = R(self._iterable).most_common
        commonality = cnt()
        return self._append(T(
            # least common
            commonality[:-2:-1][0][0],
            # most common (mode)
            cnt(1)[0][0],
            # overall commonality
            commonality,
        ))

    @memoize
    def _max(self, mx=max):
        return self._append(mx(self._iterable, key=self._identity))

    def _median(
        self, t=tee, sd=sorted, td=truediv, i=int, cnt=count, z=slicer,
    ):
        i1, i2 = t(sd(self._iterable))
        result = td(cnt(i1) - 1, 2)
        pint = i(result)
        if result % 2 == 0:
            return self._append(z(i2, pint))
        i3, i4 = t(i2)
        return self._append(td(z(i3, pint) + z(i4, pint + 1), 2))

    def _minmax(self, mn=min, mx=max, t=tee, MM=MinMax):
        i1, i2 = t(self._iterable)
        return self._append(MM(mn(i1), mx(i2)))

    def _range(self, d=deque, sd=sorted, nx=next, t=tee):
        i1, i2 = t(sd(self._iterable))
        return self._append(d(i1, maxlen=1).pop() - nx(i2))

    @memoize
    def _min(self, mn=min):
        return self._append(mn(self._iterable, key=self._identity))

    @memoize
    def _sum(self, start, floats, su=sum, fs=fsum):
        return self._append(
            fs(self._iterable) if floats else su(self._iterable, start)
        )


class _OrderMixin(local):

    @memoize
    def _group(
        self, G=GroupBy, S=StopIteration, g=groupby, nx=next, sd=sorted,
        tu=tuple,
    ):
        def grouper(call, iterable):
            try:
                it = g(sd(iterable, key=call), call)
                while 1:
                    k, v = nx(it)
                    yield G(k, tu(v))
            except S:
                pass
        return self._xtend(grouper(self._identity, self._iterable))

    def _reverse(self, S=StopIteration, nx=next, rv=reversed, tu=tuple):
        return self._xtend(rv(tu(self._iterable)))

    def _shuffle(self, l=list, sf=shuffle):
        iterable = l(self._iterable)
        sf(iterable)
        return self._xtend(iterable)

    @memoize
    def _sort(self, sd=sorted):
        return self._xtend(sd(self._iterable, key=self._identity))


class _RepeatMixin(local):

    def _combinate(self, n, cb=combinations):
        return self._xtend(cb(self._iterable, n))

    def _copy(self, dc=deepcopy, m=map):
        return self._xtend(m(dc, self._iterable))

    def _permute(self, n, pm=permutations):
        return self._xtend(pm(self._iterable, n))

    @memoize
    def _repeat(self, n, use, rt=repeat, sm=starmap, tu=tuple):
        call = self._identity
        if use:
            return self._xtend(sm(call, rt(tu(self._iterable), n)))
        return self._xtend(rt(tu(self._iterable), n))


class _MapMixin(local):

    @memoize
    def _argmap(self, curr, sm=starmap):
        call = self._worker
        if curr:
            def argmap(*args):
                return call(*(args + self._args))
            return self._xtend(sm(argmap, self._iterable))
        return self._xtend(sm(call, self._iterable))

    @memoize
    def _invoke(self, name, m=map, mc=methodcaller):
        def invoke(thing, caller=mc(name, *self._args, **self._kw)):
            read = caller(thing)
            return thing if read is None else read
        return self._xtend(m(invoke, self._iterable))

    @memoize
    def _kwargmap(self, curr, sm=starmap):
        call = self._worker
        if curr:
            def kwargmap(*params):
                args, kwargs = params
                kwargs.update(self._kw)
                return call(*(args + self._args), **kwargs)
        else:
            kwargmap = lambda x, y: call(*x, **y)
        return self._xtend(sm(kwargmap, self._iterable))

    @memoize
    def _map(self, m=map):
        return self._xtend(m(self._worker, self._iterable))

    @memoize
    def _mapping(
        self, key, value, ky=keys, it=items, vl=values, sm=starmap,
        m=map, ci=chain.from_iterable,
    ):
        call = self._identity
        if key:
            return self._xtend(m(call, ci(m(ky, self._iterable))))
        elif value:
            return self._xtend(m(call, ci(m(vl, self._iterable))))
        return self._xtend(sm(call, ci(m(it, self._iterable))))


class _FilterMixin(local):

    @memoize
    def _attrs(
        self, names, A=AttributeError, S=StopIteration, ag=attrgetter, nx=next
    ):
        def attrs(iterable, get=ag(*names)):
            try:
                while 1:
                    try:
                        yield get(nx(iterable))
                    except A:
                        pass
            except S:
                pass
        return self._xtend(attrs(self._iterable))

    @memoize
    def _duality(
        self, TF=TrueFalse, f=filter, ff=filterfalse, tu=tuple, t=tee,
    ):
        truth, false = t(self._iterable)
        call = self._test
        return self._append(TF(tu(f(call, truth)), tu(ff(call, false))))

    @memoize
    def _filter(self, false, f=filter, ff=filterfalse):
        return self._xtend((ff if false else f)(self._worker, self._iterable))

    @memoize
    def _items(
        self, key, ig=itemgetter, I=IndexError, K=KeyError, T=TypeError,
        nx=next, S=StopIteration,
    ):
        def itemz(iterable, get=ig(*key)):
            try:
                while 1:
                    try:
                        yield get(nx(iterable))
                    except (I, K, T):
                        pass
            except S:
                pass
        return self._xtend(itemz(self._iterable))

    @memoize
    def _traverse(
        self, invert, CM=ChainMap, OD=OrderedDict, S=StopIteration, cn=chain,
        d=dir, f=filter, ff=filterfalse, ga=getattr, gm=getmro, ic=isclass,
        ii=isinstance, nx=next, se=set, sn=selfname, vz=vars, it=iter,
    ):
        if self._worker is None:
            test = lambda x: not x[0].startswith('__')
        else:
            test = self._worker
        ifilter = ff if invert else f
        def members(iterable):  # @IgnorePep8
            mro = gm(iterable)
            names = it(d(iterable))
            beenthere = se()
            adder = beenthere.add
            try:
                while 1:
                    name = nx(names)
                    # yes, it's really supposed to be a tuple
                    for base in cn([iterable], mro):
                        var = vz(base)
                        if name in var:
                            obj = var[name]
                            break
                    else:
                        obj = ga(iterable, name)
                    if (name, obj) in beenthere:
                        continue
                    else:
                        adder((name, obj))
                    if ic(obj):
                        yield name, OD((k, v) for k, v in ifilter(
                            test, members(obj),
                        ))
                    else:
                        yield name, obj
            except StopIteration:
                pass
        def traverse(iterable):  # @IgnorePep8
            try:
                iterable = it(iterable)
                while 1:
                    iterator = nx(iterable)
                    chaining = CM()
                    chaining['classname'] = sn(iterator)
                    cappend = chaining.maps.append
                    for k, v in ifilter(test, members(iterator)):
                        if isinstance(v, OD):
                            v['classname'] = k
                            cappend(v)
                        else:
                            chaining[k] = v
                    yield chaining
            except StopIteration:
                pass
        return self._xtend(traverse(self._iterable))


class _ReduceMixin(local):

    def _flatten(
        self, A=AttributeError, S=StopIteration, T=TypeError, nx=next,
        st=isstring,
    ):
        def flatten(iterable):
            next_ = iterable.__iter__()
            try:
                while 1:
                    item = nx(next_)
                    try:
                        # don't recur over strings
                        if st(item):
                            yield item
                        else:
                            # do recur over other things
                            for j in flatten(item):
                                yield j
                    except (A, T):
                        # does not recur
                        yield item
            except S:
                pass
        return self._xtend(flatten(self._iterable))

    def _merge(self, ci=chain.from_iterable):
        return self._xtend(ci(self._iterable))

    @memoize
    def _reduce(self, initial, reverse, rz=reduce):
        call = self._worker
        if reverse:
            if initial is None:
                return self._append(
                    rz(lambda x, y: call(y, x), self._iterable)
                )
            return self._append(
                rz(lambda x, y: call(y, x), self._iterable, initial)
            )
        if initial is None:
            return self._append(rz(call, self._iterable))
        return self._append(rz(call, self._iterable, initial))

    def _zip(self, zip_=zip_longest):
        return self._xtend(zip_(*self._iterable))


class _SliceMixin(local):

    @memoize
    def _at(self, n, default, iz=islice, nx=next):
        return self._append(nx(iz(self._iterable, n, None), default))

    @memoize
    def _choice(self, cnt=count, iz=islice, nx=next, rr=randrange, t=tee):
        i1, i2 = t(self._iterable)
        return self._append(iz(i1, rr(0, cnt(i2)), None))

    @memoize
    def _dice(self, n, fill, zl=zip_longest):
        return self._xtend(
            zl(fillvalue=fill, *[self._iterable.__iter__()] * n)
        )

    @memoize
    def _first(self, n=0, df=deferiter, iz=islice):
        return self._xtend(iz(self._iterable, n) if n else df(self._iterable))

    def _initial(self, cnt=count, iz=islice, t=tee):
        i1, i2 = t(self._iterable)
        return self._xtend(iz(i1, cnt(i2) - 1))

    @memoize
    def _last(self, n, cnt=count, d=deque, df=deferfunc, iz=islice, t=tee):
        if n:
            i1, i2 = t(self._iterable)
            return self._xtend(iz(i1, cnt(i2) - n, None))
        return self._xtend(df(d(self._iterable, maxlen=1).pop))

    def _rest(self, iz=islice):
        return self._xtend(iz(self._iterable, 1, None))

    @memoize
    def _sample(self, n, t=tee, z=slice, rr=randrange, m=map, cnt=count):
        i1, i2 = t(self._iterable)
        length = cnt(i1)
        return self._xtend(m(lambda x: z(x, rr(0, length)), t(i2, n)))

    def _slice(self, start, stop, step, iz=islice):
        if stop and step:
            return self._xtend(iz(self._iterable, start, stop, step))
        elif stop:
            return self._xtend(iz(self._iterable, start, stop))
        return self._xtend(iz(self._iterable, start))



# -*- coding: utf-8 -*-
'''ordering test mixins'''

import platform

PYPY = platform.python_implementation() == 'PyPy'

class stooges: #@IgnorePep8
    name = 'moe'
    age = 40
class stoog2: #@IgnorePep8
    name = 'larry'
    age = 50
class stoog3: #@IgnorePep8
    name = 'curly'
    age = 60
    class stoog4: #@IgnorePep8
        name = 'beastly'
        age = 969


class stooge: #@IgnorePep8
    name = 'moe'
    age = 40
class stooge2(stooge): #@IgnorePep8
    pname = 'larry'
    page = 50
class stooge3(stoog2): #@IgnorePep8
    aname = 'curly'
    rage = 60
    class stooge4(stooge): #@IgnorePep8
        kname = 'beastly'
        mage = 969


class MathMixin(object):

    def test_pipe(self):
        one = self.mclass(10, 5, 100, 2, 1000)
        two = self.pipe()
        test = one.minmax().pipe(two).merge().back().min().get()
        self.assertEqual(test, 2, test)
        test = one.original().minmax().pipe(two).merge().back().max().get()
        self.assertEqual(test, 1000, test)
        test = one.original().minmax().pipe(two).merge().back().sum().get()
        self.assertEqual(test, 1002, test)

    def test_average(self):
        self.assertEqual(
            self.mclass(10, 40, 45).average().get(), 31.666666666666668,
        )

    def test_count(self):
        common = self.mclass(11, 3, 5, 11, 7, 3, 11).count().get()
        self.assertEqual(common.overall, [(11, 3), (3, 2), (5, 1), (7, 1)])
        # most common
        self.assertEqual(common.most, 11)
        # least common
        self.assertEqual(common.least, 7)

    def test_max(self):
        from stuf import stuf
        stooge = [
            stuf(name='moe', age=40),
            stuf(name='larry', age=50),
            stuf(name='curly', age=60),
        ]
        self.assertEqual(self.mclass(1, 2, 4).max().get(), 4)
        self.assertEqual(
            stuf(self.mclass(*stooge).worker(lambda x: x.age).max().get()),
            stuf(name='curly', age=60),
        )

    def test_median(self):
        self.assertEqual(self.mclass(4, 5, 7, 2, 1).median().get(), 4)
        self.assertEqual(self.mclass(4, 5, 7, 2, 1, 8).median().get(), 4.5)

    def test_min(self):
        self.assertEqual(self.mclass(10, 5, 100, 2, 1000).min().get(), 2)
        self.assertEqual(
            self.mclass(10, 5, 100, 2, 1000).worker(
            lambda x: x % 100 == 0
            ).min().get(),
            10,
        )

    def test_minmax(self):
        self.assertEqual(self.mclass(1, 2, 4).minmax().get(), (1, 4))
        self.assertEqual(
            self.mclass(10, 5, 100, 2, 1000).minmax().get(), (2, 1000),
        )

    def test_range(self):
        self.assertEqual(self.mclass(3, 5, 7, 3, 11).range().get(), 8)

    def test_sum(self):
        self.assertEqual(self.mclass(1, 2, 3).sum().get(), 6)
        self.assertEqual(self.mclass(1, 2, 3).sum(1).get(), 7)
        self.assertEqual(
            self.mclass(
                .1, .1, .1, .1, .1, .1, .1, .1, .1, .1
            ).sum(precision=True).get(),
            1.0,
        )


class CmpMixin(object):

    def test_all(self):
        from operator import truth
        self.assertFalse(
            self.mclass(True, 1, None, 'yes').worker(truth).all().get()
        )
        self.assertTrue(
            self.mclass(2, 4, 6, 8).worker(lambda x: x % 2 == 0).all().get()
        )

    def test_any(self):
        self.assertTrue(
            self.mclass(None, 0, 'yes', False).worker(bool).any().get()
        )
        self.assertTrue(
            self.mclass(1, 4, 5, 9).worker(lambda x: x % 2 == 0).any().get()
        )

    def test_difference(self):
        self.assertEqual(
            self.mclass(
                [1, 2, 3, 4, 5], [5, 2, 10], [10, 11, 2]
            ).difference().get(),
            [1, 3, 4],
        )
        self.assertEqual(
            self.mclass(
                [1, 3, 4, 5], [5, 2, 10], [10, 11, 2]
            ).difference(True).get(),
            [3, 1, 11, 4] if PYPY else [1, 3, 4, 11]
        )

    def test_intersection(self):
        self.assertEqual(
            self.mclass(
                [1, 2, 3], [101, 2, 1, 10], [2, 1]
            ).intersection().get(), [1, 2],
        )

    def test_union(self):
        self.assertEqual(
            self.mclass([1, 2, 3], [101, 2, 1, 10], [2, 1]).union().get(),
            [10, 1, 2, 3, 101] if PYPY else [1, 10, 3, 2, 101],
        )

    def test_unique(self):
        self.assertEqual(
            self.mclass(1, 2, 1, 3, 1, 4).unique().get(), [1, 2, 3, 4],
        )
        self.assertEqual(
            self.mclass(1, 2, 1, 3, 1, 4).worker(round).unique().get(),
            [1, 2, 3, 4],
        )


class OrderMixin(object):

    def test_shuffle(self):
        self.assertEqual(
            len(self.mclass(1, 2, 3, 4, 5, 6).shuffle()),
            len([5, 4, 6, 3, 1, 2]),
        )

    def test_group(self,):
        self.assertEqual(
            self.mclass(1.3, 2.1, 2.4).group().get(),
            [(1.3, (1.3,)), (2.1, (2.1,)), (2.4, (2.4,))],
        )
        from math import floor
        self.assertEqual(
            self.mclass(1.3, 2.1, 2.4).worker(floor).group().get(),
            [(1.0, (1.3,)), (2.0, (2.1, 2.4))]
        )

    def test_combo(self):
        self.assertEqual(
            self.mclass(5, 4, 3, 2, 1).reverse().sort().get(), [1, 2, 3, 4, 5]
        )

    def test_reverse(self):
        self.assertEqual(
            self.mclass(5, 4, 3, 2, 1).reverse().get(), [1, 2, 3, 4, 5],
        )

    def test_sort(self):
        from math import sin
        self.assertEqual(
            self.mclass(1, 2, 3, 4, 5, 6).worker(sin).sort().get(),
            [5, 4, 6, 3, 1, 2],
        )
        self.assertEqual(
            self.mclass(4, 6, 65, 3, 63, 2, 4).sort().get(),
            [2, 3, 4, 4, 6, 63, 65],
        )


class FilterMixin(object):

    def test_pattern(self):
        test = self.mclass(
            'This is the first test',
            'This is the second test',
            'This is the third test',
        ).pattern('{} first {}')
        self.assertEqual(
            test.filter().get(), 'This is the first test'
        )
        self.assertEqual(
            test.original().pattern(
                '. third .', type='regex'
            ).filter().get(), 'This is the third test'
        )
        self.assertEqual(
            test.original().pattern(
                '* second *', type='glob'
            ).filter().get(), 'This is the second test'
        )

    def test_traverse(self):
        self.maxDiff = None
        from stuf.collects import ChainMap, OrderedDict
        get = self.mclass(stoog3).traverse().get()
        self.assertEqual(
            get,
            ChainMap(OrderedDict([
                ('classname', 'stoog3'), ('age', 60), ('name', 'curly')]),
                OrderedDict([
                ('age', 969), ('name', 'beastly'), ('classname', 'stoog4')]))
        )
        def test(x): #@IgnorePep8
            if x[0] == 'name':
                return True
            elif x[0].startswith('__'):
                return True
            return False
        self.assertEqual(
            self.mclass(
                stooges, stoog2, stoog3
            ).worker(test).traverse(True).get(),
            [ChainMap(OrderedDict([('classname', 'stooges'), ('age', 40)])),
            ChainMap(OrderedDict([('classname', 'stoog2'), ('age', 50)])),
            ChainMap(
                OrderedDict([('classname', 'stoog3'), ('age', 60)]),
                OrderedDict([('classname', 'stoog4'), ('age', 969)])
            )],
        )
        get = self.mclass(stooge, stooge2, stooge3).traverse().get()
        self.assertEqual(
            get,
            [ChainMap(OrderedDict([
                ('classname', 'stooge'), ('age', 40), ('name', 'moe')])),
            ChainMap(
                OrderedDict([
                    ('classname', 'stooge'), ('age', 40), ('name', 'moe'),
                    ('classname', 'stooge2'), ('page', 50), ('pname', 'larry')
                ])
            ),
            ChainMap(OrderedDict([
                ('classname', 'stooge3'), ('age', 50), ('aname', 'curly'),
                ('name', 'larry'), ('rage', 60)]), OrderedDict([('age', 40),
                ('kname', 'beastly'), ('mage', 969), ('name', 'moe'),
                ('classname', 'stooge4')]))
    ])

    def test_attrs(self):
        from stuf import stuf
        stooge = [
            stuf(name='moe', age=40),
            stuf(name='larry', age=50),
            stuf(name='curly', age=60)
        ]
        self.assertEqual(
            self.mclass(*stooge).attrs('name').get(),
            ['moe', 'larry', 'curly'],
        )
        self.assertEqual(
            self.mclass(*stooge).attrs('name', 'age').get(),
            [('moe', 40), ('larry', 50), ('curly', 60)],
        )
        self.assertEqual(
            self.mclass(*stooge).attrs('place').get(), [],
        )

    def test_items(self):
        from stuf import stuf
        stooge = [
            stuf(name='moe', age=40),
            stuf(name='larry', age=50),
            stuf(name='curly', age=60)
        ]
        self.assertEqual(
            self.mclass(*stooge).items('name').get(),
            ['moe', 'larry', 'curly'],
        )
        self.assertEqual(
            self.mclass(*stooge).items('name', 'age').get(),
            [('moe', 40), ('larry', 50), ('curly', 60)],
        )
        stooge = [['moe', 40], ['larry', 50], ['curly', 60]]
        self.assertEqual(
            self.mclass(*stooge).items(0).get(), ['moe', 'larry', 'curly'],
        )
        self.assertEqual(self.mclass(*stooge).items(1).get(), [40, 50, 60])
        self.assertEqual(self.mclass(*stooge).items('place').get(), [])

    def test_filter(self):
        self.assertEqual(
            self.mclass(1, 2, 3, 4, 5, 6).worker(
                lambda x: x % 2 == 0
            ).filter(invert=True).get(), [1, 3, 5]
        )
        self.assertEqual(
            self.mclass(1, 2, 3, 4, 5, 6).worker(
                lambda x: x % 2 == 0
            ).filter().get(), [2, 4, 6]
        )

    def test_duality(self):
        self.assertEqual(
            self.mclass(1, 2, 3, 4, 5, 6).worker(
                lambda x: x % 2 == 0
            ).duality().get(),
            ((2, 4, 6), (1, 3, 5))
        )


class SliceMixin(object):

    def test_dice(self):
        self.assertEqual(
            self.mclass(
                'moe', 'larry', 'curly', 30, 40, 50, True
            ).dice(2, 'x').get(),
            [('moe', 'larry'), ('curly', 30), (40, 50), (True, 'x')]
        )

    def test_first(self):
        self.assertEqual(self.mclass(5, 4, 3, 2, 1).first().get(), 5)
        self.assertEqual(self.mclass(5, 4, 3, 2, 1).first(2).get(), [5, 4])

    def test_combo(self):
        self.assertEqual(self.mclass(
            5, 4, 3, 2, 1).initial().rest().slice(1, 2).last().get(), 3,
        )

    def test_index(self):
        self.assertEqual(self.mclass(5, 4, 3, 2, 1).at(2).get(), 3)
        self.assertEqual(self.mclass(5, 4, 3, 2, 1).at(10, 11).get(), 11)

    def test_slice(self):
        self.assertEqual(self.mclass(5, 4, 3, 2, 1).slice(2).get(), [5, 4])
        self.assertEqual(
            self.mclass(5, 4, 3, 2, 1).slice(2, 4).get(), [3, 2]
        )
        self.assertEqual(
            self.mclass(5, 4, 3, 2, 1).slice(2, 4, 2).get(), 3
        )

    def test_last(self):
        self.assertEqual(self.mclass(5, 4, 3, 2, 1).last().get(), 1)
        self.assertEqual(self.mclass(5, 4, 3, 2, 1).last(2).get(), [2, 1])

    def test_initial(self):
        self.assertEqual(
            self.mclass(5, 4, 3, 2, 1).initial().get(), [5, 4, 3, 2]
        )

    def test_rest(self):
        self.assertEqual(
            self.mclass(5, 4, 3, 2, 1).rest().get(), [4, 3, 2, 1],
        )

    def test_choice(self):
        self.assertEqual(
            len(list(self.mclass(1, 2, 3, 4, 5, 6).choice())), 1,
        )

    def test_sample(self):
        self.assertEqual(
            len(self.mclass(1, 2, 3, 4, 5, 6).sample(3).get()), 3,
        )


class ReduceMixin(object):

    def test_flatten(self):
        self.assertEqual(
            self.mclass([[1, [2], [3, [[4]]]], 'here']).flatten().get(),
            [1, 2, 3, 4, 'here'],
        )

    def test_merge(self):
        self.assertEqual(
            self.mclass(
                ['moe', 'larry', 'curly'], [30, 40, 50], [True, False, False]
            ).merge().get(),
            ['moe', 'larry', 'curly', 30, 40, 50, True, False, False],
        )

    def test_reduce(self):
        self.assertEqual(
            self.mclass(1, 2, 3).worker(lambda x, y: x + y).reduce().get(),
            6,
        )
        self.assertEqual(
            self.mclass(1, 2, 3).worker(lambda x, y: x + y).reduce(1).get(),
            7,
        )
        self.assertEqual(
            self.mclass([0, 1], [2, 3], [4, 5]).worker(
                lambda x, y: x + y
            ).reduce(reverse=True).get(), [4, 5, 2, 3, 0, 1],
        )
        self.assertEqual(
            self.mclass([0, 1], [2, 3], [4, 5]).worker(
                lambda x, y: x + y
            ).reduce([0, 0], True).get(), [4, 5, 2, 3, 0, 1, 0, 0],
        )

    def test_zip(self):
        # auto
        self.assertEqual(
            self.mclass(
                ['moe', 'larry', 'curly'], [30, 40, 50], [True, False, False]
            ).zip().get(),
            [('moe', 30, True), ('larry', 40, False), ('curly', 50, False)],
        )


class RepeatMixin(object):

    def test_repeat(self):
        def test(*args): #@IgnorePep8
            return list(args)
        self.assertEqual(
            self.mclass(40, 50, 60).repeat(3).get(),
            [(40, 50, 60), (40, 50, 60), (40, 50, 60)],
        )
        self.assertEqual(
            self.mclass(40, 50, 60).worker(test).repeat(3, True).get(),
            [[40, 50, 60], [40, 50, 60], [40, 50, 60]],
        )

    def test_copy(self):
        testlist = [[1, [2, 3]], [4, [5, 6]]]
        newlist = self.mclass(testlist).copy().get()
        self.assertFalse(newlist is testlist)
        self.assertListEqual(newlist, testlist)
        self.assertFalse(newlist[0] is testlist[0])
        self.assertListEqual(newlist[0], testlist[0])
        self.assertFalse(newlist[1] is testlist[1])
        self.assertListEqual(newlist[1], testlist[1])

    def test_permutations(self):
        self.assertEqual(
            self.mclass(40, 50, 60).permutate(2).get(),
            [(40, 50), (40, 60), (50, 40), (50, 60), (60, 40), (60, 50)],
        )

    def test_combination(self):
        self.assertEqual(
            self.mclass(40, 50, 60).combinate(2).get(),
            [(40, 50), (40, 60), (50, 60)],
        )


class MapMixin(object):

    def test_factory(self):
        from stuf import stuf
        thing = self.mclass(
            [('a', 1), ('b', 2), ('c', 3)], [('a', 1), ('b', 2), ('c', 3)]
        ).worker(stuf).map().get()
        self.assertEqual(
            thing, [stuf(a=1, b=2, c=3), stuf(a=1, b=2, c=3)]
        )

    def test_kwargmap(self):
        def test(*args, **kw):
            return sum(args) * sum(kw.values())
        self.assertEqual(
            self.mclass(
                ((1, 2), {'a': 2}), ((2, 3), {'a': 2}), ((3, 4), {'a': 2})
            ).worker(test).kwargmap().get(),
            [6, 10, 14],
        )
        self.assertEqual(
            self.mclass(
                ((1, 2), {'a': 2}), ((2, 3), {'a': 2}), ((3, 4), {'a': 2})
            ).worker(test).params(
                1, 2, 3, b=5, w=10, y=13
            ).kwargmap(True).get(),
            [270, 330, 390],
        )
        self.assertEqual(
            self.mclass(
                ((1, 2), {'a': 2}), ((2, 3), {'a': 2}), ((3, 4), {'a': 2})
            ).apply(
                test, 1, 2, 3, b=5, w=10, y=13
            ).kwargmap(True).get(),
            [270, 330, 390],
        )

    def test_argmap(self):
        self.assertEqual(
            self.mclass(
                (1, 2), (2, 3), (3, 4)
            ).worker(lambda x, y: x * y).argmap().get(),
            [2, 6, 12],
        )
        self.assertEqual(
            self.mclass((1, 2), (2, 3), (3, 4)).worker(
                lambda x, y, z, a, b: x * y * z * a * b
            ).params(7, 8, 9).argmap(True).get(),
            [1008, 3024, 6048],
        )
        self.assertEqual(
            self.mclass((1, 2), (2, 3), (3, 4)).apply(
                lambda x, y, z, a, b: x * y * z * a * b, 7, 8, 9
            ).argmap(True).get(),
            [1008, 3024, 6048],
        )

    def test_map(self):
        self.assertEqual(
            self.mclass(1, 2, 3).worker(lambda x: x * 3).map().get(),
            [3, 6, 9],
        )

    def test_invoke(self):
        self.assertEqual(
            self.mclass(
                [5, 1, 7], [3, 2, 1]
            ).params(1).invoke('index').get(),
            [1, 2],
        )
        self.assertEqual(
            self.mclass([5, 1, 7], [3, 2, 1]).invoke('sort').get(),
            [[1, 5, 7], [1, 2, 3]],
        )

    def test_mapping(self):
        self.assertEqual(
            self.mclass(
                dict([(1, 2), (2, 3), (3, 4)]), dict([(1, 2), (2, 3), (3, 4)])
            ).mapping(True).get(), [1, 2, 3, 1, 2, 3],
        )
        self.assertEqual(
            self.mclass(
                dict([(1, 2), (2, 3), (3, 4)]), dict([(1, 2), (2, 3), (3, 4)])
            ).mapping(values=True).get(),
            [2, 3, 4, 2, 3, 4],
        )
        self.assertEqual(
            self.mclass(
                dict([(1, 2), (2, 3), (3, 4)]), dict([(1, 2), (2, 3), (3, 4)])
            ).worker(lambda x, y: x * y).mapping().get(),
            [2, 6, 12, 2, 6, 12],
        )


class Mixin(object):

    def test_repr(self):
        from stuf.six import strings
        self.assertIsInstance(
            self.mclass([1, 2, 3, 4, 5, 6]).__repr__(), strings,
        )

    def test_append(self):
        self.assertEqual(self.mclass().append('foo').peek(), 'foo')
        self.assertListEqual(
            self.mclass().append(1, 2, 3, 4, 5, 6).peek(),
            [1, 2, 3, 4, 5, 6],
        )

    def test_prepend(self):
        self.assertEqual(self.mclass().prepend('foo').peek(), 'foo')
        self.assertListEqual(
            self.mclass().prepend(1, 2, 3, 4, 5, 6).peek(), [1, 2, 3, 4, 5, 6]
        )

    def test_undo(self):
        queue = self.mclass(1, 2, 3).prepend(1, 2, 3, 4, 5, 6)
        self.assertEqual(queue.peek(), [1, 2, 3, 4, 5, 6, 1, 2, 3])
        queue.append(1).undo()
        self.assertEqual(queue.peek(), [1, 2, 3, 4, 5, 6, 1, 2, 3])
        queue.append(1).append(2).undo()
        self.assertEqual(queue.peek(), [1, 2, 3, 4, 5, 6, 1, 2, 3, 1])
        queue.append(1).append(2).undo(2)
        self.assertEqual(queue.peek(), [1, 2, 3, 4, 5, 6, 1, 2, 3, 1])
        queue.snapshot().append(1).append(2).baseline()
        self.assertEqual(queue.peek(), [1, 2, 3, 4, 5, 6, 1, 2, 3, 1])
        queue.original()
        self.assertEqual(queue.peek(), [1, 2, 3])

    def test_wrap(self):
        self.assertIsInstance(
            self.mclass(1, 2, 3, 4, 5, 6).wrap(tuple).peek(), tuple,
        )
        self.assertTupleEqual(
            self.mclass(1, 2, 3, 4, 5, 6).wrap(tuple).peek(),
            (1, 2, 3, 4, 5, 6),
        )

    def test_ascii(self):
        from stuf.six import u, b
        self.assertEqual(
            self.mclass(
                [1], True, r't', b('i'), u('g'), None, (1,)
            ).ascii().oneach().peek(),
            [b('[1]'), b('True'), b('t'), b('i'), b('g'), b('None'), b('(1,)')]
        )

    def test_bytes(self):
        from stuf.six import u, b
        self.assertEqual(
            self.mclass(
                [1], True, r't', b('i'), u('g'), None, (1,)
            ).bytes().oneach().peek(),
            [b('[1]'), b('True'), b('t'), b('i'), b('g'), b('None'), b('(1,)')]
        )

    def test_unicode(self):
        from stuf.six import u, b
        self.assertEqual(
            self.mclass(
                [1], True, r't', b('i'), u('g'), None, (1,)
            ).unicode().oneach().peek(),
            [u('[1]'), u('True'), u('t'), u('i'), u('g'), u('None'), u('(1,)')]
        )



# -*- coding: utf-8 -*-
'''lazy knife tests'''

from stuf.six import unittest

from tests.mixins import (
    Mixin, MapMixin, RepeatMixin, ReduceMixin, SliceMixin, FilterMixin,
    OrderMixin, MathMixin, CmpMixin)


class TestLazy(
    unittest.TestCase, Mixin, CmpMixin, MapMixin, ReduceMixin, OrderMixin,
    SliceMixin, RepeatMixin, MathMixin, FilterMixin,
):

    def setUp(self):
        from knife import lazyknife
        self.mclass = lazyknife
        self.pipe = lazyknife


class TestCompare(unittest.TestCase, Mixin, CmpMixin):

    def setUp(self):
        from knife.lazy import cmpknife
        self.mclass = cmpknife


class TestFilter(unittest.TestCase, Mixin, FilterMixin):

    def setUp(self):
        self.maxDiff = None
        from knife.lazy import filterknife
        self.mclass = filterknife


class TestMap(unittest.TestCase, Mixin, MapMixin):

    def setUp(self):
        from knife.lazy import mapknife
        self.mclass = mapknife


class TestMath(unittest.TestCase, Mixin, MathMixin):

    def setUp(self):
        from knife.lazy import mathknife, reduceknife
        self.mclass = mathknife
        self.pipe = reduceknife


class TestOrder(unittest.TestCase, Mixin, OrderMixin):

    def setUp(self):
        from knife.lazy import orderknife
        self.mclass = orderknife


class TestRepeat(unittest.TestCase, Mixin, RepeatMixin):

    def setUp(self):
        from knife.lazy import repeatknife
        self.mclass = repeatknife


class TestReduce(unittest.TestCase, Mixin, ReduceMixin):

    def setUp(self):
        from knife.lazy import reduceknife
        self.mclass = reduceknife


class TestSlice(unittest.TestCase, Mixin, SliceMixin):

    def setUp(self):
        from knife.lazy import sliceknife
        self.mclass = sliceknife

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
'''Things go in. Things happen. Things come out.'''

from knife.lazy import lazyknife
from knife.active import activeknife

knife = activeknife
__ = lazyknife

__all__ = ('knife', 'activeknife', 'lazyknife', '__')

__version__ = (0, 5, 10)



# -*- coding: utf-8 -*-
'''lazily evaluated knives'''

from threading import local
from itertools import tee, chain
from contextlib import contextmanager

from stuf.deep import clsname
from stuf.iterable import count


class _LazyMixin(local):

    '''Lazy knife mixin.'''

    def __init__(self, *things, **kw):
        '''
        :argument things: incoming things
        :keyword integer snapshots: snapshots to keep (default: ``5``)
        '''
        incoming = (
            (things[0],).__iter__() if len(things) == 1 else things.__iter__()
        )
        super(_LazyMixin, self).__init__(incoming, ().__iter__(), **kw)
        # working things
        self._work = ().__iter__()
        # holding things
        self._hold = ().__iter__()

    @property
    @contextmanager
    def _chain(self, tee_=tee):
        # take snapshot
        self._in, snapshot = tee_(self._in)
        # rebalance incoming with outcoming
        if self._history:
            self._in, self._out = tee_(self._out)
        # make snapshot original snapshot?
        else:
            self._original = snapshot
        # place snapshot at beginning of snapshot stack
        self._history.appendleft(snapshot)
        # move incoming things to working things
        work, self._in = tee_(self._in)
        self._work = work
        yield
        # extend outgoing things with holding things
        self._out = self._hold
        # clear working things
        del self._work
        self._work = ().__iter__()
        # clear holding things
        del self._hold
        self._hold = ().__iter__()

    @property
    def _iterable(self):
        # iterable derived from link in chain
        return self._work

    def _xtend(self, things, chain_=chain):
        # place things after holding things
        self._hold = chain_(things, self._hold)
        return self

    def _append(self, things, chain_=chain):
        # append thing after other holding things
        self._hold = chain_(self._hold, (things,).__iter__())
        return self

    def _prependit(self, things, tee_=tee, chain_=chain):
        # take snapshot
        self._in, snapshot = tee_(self._in)
        # make snapshot original snapshot?
        if self._original is None:
            self._original = snapshot
        # place snapshot at beginning of snapshot stack
        self._history.appendleft(snapshot)
        # place things before other incoming things
        self._in = chain_(things, self._in)
        return self

    def _appendit(self, things, tee_=tee, chain_=chain):
        # take snapshot
        self._in, snapshot = tee_(self._in)
        # make snapshot original snapshot?
        if self._original is None:
            self._original = snapshot
        # place snapshot at beginning of snapshot stack
        self._history.appendleft(snapshot)
        # place things before other incoming things
        self._in = chain_(self._in, things)
        return self

    def _pipeit(self, knife):
        knife.clear()
        knife._history.clear()
        knife._history.extend(self._history)
        knife._original = self._original
        knife._baseline = self._baseline
        knife._out = self._out
        knife._worker = self._worker
        knife._args = self._args
        knife._kw = self._kw
        knife._wrapper = self._wrapper
        knife._pipe = self
        return knife

    def _unpipeit(self):
        piped = self._pipe
        piped.clear()
        piped._history.clear()
        piped._history.extend(self._history)
        piped._original = self._original
        piped._baseline = self._baseline
        piped._out = self._out
        piped._worker = self._worker
        piped._args = self._args
        piped._kw = self._kw
        piped._wrapper = self._wrapper
        return piped

    def _repr(self, tee_=tee, l=list, clsname_=clsname):
        # object representation
        self._in, in2 = tee_(self._in)
        self._out, out2 = tee_(self._out)
        self._work, work2 = tee_(self._work)
        self._hold, hold2 = tee_(self._hold)
        return self._REPR.format(
            self.__module__,
            clsname_(self),
            l(in2),
            l(work2),
            l(hold2),
            l(out2),
        )

    def _len(self, tee_=tee, count_=count):
        # length of incoming things
        self._in, incoming = tee_(self._in)
        return count_(incoming)


class _OutMixin(_LazyMixin):

    '''Lazy output mixin.'''

    def _undo(self, snapshot=0):
        # clear everything
        self.clear()
        # if specified, use a specific snapshot
        if snapshot:
            self._history.rotate(-(snapshot - 1))
        self._in = self._history.popleft()
        # clear outgoing things
        del self._out
        self._out = ().__iter__()
        return self

    def _snapshot(self, tee_=tee):
        # take baseline snapshot of incoming things
        self._in, self._baseline = tee_(self._in)
        return self

    def _rollback(self, tee_=tee):
        # clear everything
        self.clear()
        # clear snapshots
        self._clearsp()
        # revert to baseline snapshot of incoming things
        self._in, self._baseline = tee_(self._baseline)
        return self

    def _revert(self, tee_=tee):
        # clear everything
        self.clear()
        # clear snapshots
        self._clearsp()
        # clear baseline
        self._baseline = None
        # restore original snapshot of incoming things
        self._in, self._original = tee_(self._original)
        return self

    def _clear(self, list_=list):
        # clear worker
        self._worker = None
        # clear worker positional arguments
        self._args = ()
        # clear worker keyword arguments
        self._kw = {}
        # revert to default iterable wrapper
        self._wrapper = list_
        # clear pipe
        self._pipe = None
        # clear incoming things
        del self._in
        self._in = ().__iter__()
        # clear working things
        del self._work
        self._work = ().__iter__()
        # clear holding things
        del self._hold
        self._hold = ().__iter__()
        # clear outgoing things
        del self._out
        self._out = ().__iter__()
        return self

    def _iterate(self, tee_=tee):
        self._out, outs = tee_(self._out)
        return outs

    def _peek(self, tee_=tee, list_=list, count_=count):
        tell, self._in, out = tee_(self._in, 3)
        wrap = self._wrapper
        value = list_(wrap(i) for i in out) if self._each else wrap(out)
        # reset each flag
        self._each = False
        # reset wrapper
        self._wrapper = list_
        return value[0] if count_(tell) == 1 else value

    def _get(self, tee_=tee, list_=list, count_=count):
        tell, self._out, out = tee_(self._out, 3)
        wrap = self._wrapper
        value = list_(wrap(i) for i in out) if self._each else wrap(out)
        # reset each flag
        self._each = False
        # reset wrapper
        self._wrapper = list_
        return value[0] if count_(tell) == 1 else value



# -*- coding: utf-8 -*-
'''knife mixins.'''

from threading import local


class CmpMixin(local):

    '''Comparing knife mixin.'''

    def all(self):
        '''
        Discover if :meth:`worker` is :data:`True` for **all** incoming things.

        :return: :func:`bool`

        :rtype: :obj:`knife` :term:`object`

        >>> from knife import __
        >>> __(2, 4, 6, 8).worker(lambda x: x % 2 == 0).all().get()
        True

        .. seealso::

          :func:`~all`
            function in Python standard library

          `all <http://documentcloud.github.com/underscore/#all>`_
            function in Underscore.js

          `all <http://mirven.github.com/underscore.lua/#all>`_
            function in Underscore.lua

          `all <http://vti.github.com/underscore-perl/#all>`_
            function in Underscore.perl

          `all <http://brianhaveri.github.com/Underscore.php/#all>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._all()

    def any(self):
        '''
        Discover if :meth:`worker` is :data:`True` for **any** incoming thing.

        :return: :func:`bool`

        :rtype: :obj:`knife` :term:`object`

        >>> __(1, 4, 5, 9).worker(lambda x: x % 2 == 0).any().get()
        True

        .. seealso::

          :func:`~any`
            function in Python standard library

          `any <http://documentcloud.github.com/underscore/#any>`_
            function in Underscore.js

          `any <http://mirven.github.com/underscore.lua/#any>`_
            function in Underscore.lua

          `any <http://vti.github.com/underscore-perl/#ant>`_
            function in Underscore.perl

          `any <http://brianhaveri.github.com/Underscore.php/#any>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._any()

    def difference(self, symmetric=False):
        '''
        Discover difference within a series of :term:`iterable` incoming
        things.

        :keyword bool symmetric: return symmetric difference

        :return: :func:`list`

        :rtype: :obj:`knife` :term:`object`

        >>> # default behavior
        >>> test = __([1, 2, 3, 4, 5], [5, 2, 10], [10, 11, 2])
        >>> test.difference().get()
        [1, 3, 4]
        >>> # symmetric difference
        >>> test.original().difference(symmetric=True).get()
        [1, 2, 3, 4, 11]

        .. seealso::

          :meth:`set.difference`
            function in Python standard library

          :meth:`set.symmetric_difference`
            function in Python standard library

          `difference <http://documentcloud.github.com/underscore/#difference>`_
            function in Underscore.js

          `difference <http://vti.github.com/underscore-perl/#difference>`_
            function in Underscore.perl

          `difference <http://brianhaveri.github.com/Underscore.php/#difference>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._difference(symmetric)

    def intersection(self):
        '''
        Discover intersection within a series of :term:`iterable` incoming
        things.

        :return: :func:`list`

        :rtype: :obj:`knife` :term:`object`

        >>> __([1, 2, 3], [101, 2, 1, 10], [2, 1]).intersection().get()
        [1, 2]

        .. seealso::

          :meth:`set.intersection`
            function in Python standard library

          `intersection <http://documentcloud.github.com/underscore/#intersection>`_
            function in Underscore.js

          `intersection <http://vti.github.com/underscore-perl/#intersection>`_
            function in Underscore.perl

          `intersection <http://brianhaveri.github.com/Underscore.php/#intersection>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._intersection()

    def union(self):
        '''
        Discover union within a series of :term:`iterable` incoming things.

        :return: :func:`list`

        :rtype: :obj:`knife` :term:`object`

        >>> __([1, 2, 3], [101, 2, 1, 10], [2, 1]).union().get()
        [1, 10, 3, 2, 101]

        .. seealso::

          :meth:`set.union`
            function in Python standard library

          `union <http://documentcloud.github.com/underscore/#union>`_
            function in Underscore.js

          `union <http://vti.github.com/underscore-perl/#union>`_
            function in Underscore.perl

          `union <http://brianhaveri.github.com/Underscore.php/#union>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._union()

    def unique(self):
        '''
        Discover unique incoming things.

        :rtype: :obj:`knife` :term:`object`

        >>> # default behavior
        >>> __(1, 2, 1, 3, 1, 4).unique().get()
        [1, 2, 3, 4]
        >>> # using worker as key function
        >>> __(1, 2, 1, 3, 1, 4).worker(round).unique().get()
        [1.0, 2.0, 3.0, 4.0]

        .. seealso::

          `uniq <http://documentcloud.github.com/underscore/#uniq>`_
            function in Underscore.js

          `uniq <http://vti.github.com/underscore-perl/#uniq>`_
            function in Underscore.perl

          `uniq <http://brianhaveri.github.com/Underscore.php/#uniq>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._unique()


class MathMixin(local):

    '''Mathing knife mixin.'''

    def average(self):
        '''
        Discover average value among incoming things.

        :return: a number

        :rtype: :obj:`knife` :term:`object`

        >>> from knife import __
        >>> __(10, 40, 45).average().get()
        31.666666666666668
        '''
        with self._chain:
            return self._average()

    def count(self):
        '''
        Discover how common each incoming thing is and the overall count of
        each incoming thing.

        :return: Collects :func:`~collections.namedtuple` ``Count(least=int,
          most=int, overall=[(thing1, int), (thing2, int), ...])``

        :rtype: :obj:`knife` :term:`object`

        >>> common = __(11, 3, 5, 11, 7, 3, 5, 11).count().get()
        >>> # least common thing
        >>> common.least
        7
        >>> # most common thing
        >>> common.most
        11
        >>> # total count for every thing
        >>> common.overall
        [(11, 3), (3, 2), (5, 2), (7, 1)]
        '''
        with self._chain:
            return self._count()

    def max(self):
        '''
        Discover the maximum value among incoming things using :meth:`worker`
        as the :term:`key function`.

        :rtype: :obj:`knife` :term:`object`

        >>> # default behavior
        >>> __(1, 2, 4).max().get()
        4
        >>> stooges = (
        ...    {'name': 'moe', 'age': 40},
        ...    {'name': 'larry', 'age': 50},
        ...    {'name': 'curly', 'age': 60},
        ... )
        >>> # using worker as key function
        >>> __(*stooges).worker(lambda x: x['age']).max().get()
        {'age': 60, 'name': 'curly'}

        .. seealso::

          :func:`~max`
            function in Python standard library

          `max <http://documentcloud.github.com/underscore/#max>`_
            function in Underscore.js

          `max <http://mirven.github.com/underscore.lua/#max>`_
            function in Underscore.lua

          `max <http://vti.github.com/underscore-perl/#max>`_
            function in Underscore.perl

          `max <http://brianhaveri.github.com/Underscore.php/#max>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._max()

    def median(self):
        '''
        Discover the median value among incoming things.

        :return: a number

        :rtype: :obj:`knife` :term:`object`

        >>> __(4, 5, 7, 2, 1).median().get()
        4
        >>> __(4, 5, 7, 2, 1, 8).median().get()
        4.5
        '''
        with self._chain:
            return self._median()

    def min(self):
        '''
        Discover the minimum value among incoming things using :meth:`worker`
        as the :term:`key function`.

        :rtype: :obj:`knife` :term:`object`

        >>> test = __(10, 5, 100, 2, 1000)
        >>> test.min().get()
        2
        >>> test.original().worker(lambda x: x % 100 == 0).min().get()
        10

        .. seealso::

          :func:`~min`
            function in Python standard library

          `min <http://documentcloud.github.com/underscore/#min>`_
            function in Underscore.js

          `min <http://mirven.github.com/underscore.lua/#min>`_
            function in Underscore.lua

          `min <http://vti.github.com/underscore-perl/#min>`_
            function in Underscore.perl

          `min <http://brianhaveri.github.com/Underscore.php/#min>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._min()

    def minmax(self):
        '''
        Discover the minimum and maximum values among incoming things.

        :return:  :func:`~collections.namedtuple` ``MinMAx(min=value,
          max=value)``.

        :rtype: :obj:`knife` :term:`object`

        >>> minmax = __(1, 2, 4).minmax().get()
        >>> minmax.min
        1
        >>> minmax.max
        4
        '''
        with self._chain:
            return self._minmax()

    def range(self):
        '''
        Discover the length of the smallest interval that can contain the
        value of every incoming thing.

        :return: a number

        :rtype: :obj:`knife` :term:`object`

        >>> __(3, 5, 7, 3, 11).range().get()
        8
        '''
        with self._chain:
            return self._range()

    def sum(self, start=0, precision=False):
        '''
        Discover the total value of adding `start` and incoming things
        together.

        :keyword start: starting number
        :type start: :func:`int` or :func:`float`

        :keyword bool precision: add floats with extended precision

        :rtype: :obj:`knife` :term:`object`

        >>> # default behavior
        >>> __(1, 2, 3).sum().get()
        6
        >>> # with a starting mumber
        >>> __(1, 2, 3).sum(start=1).get()
        7
        >>> # add floating points with extended precision
        >>> __(.1, .1, .1, .1, .1, .1, .1, .1).sum(precision=True).get()
        0.8

        .. seealso::

          :func:`sum`
            function in Python standard library
        '''
        with self._chain:
            return self._sum(start, precision)


class OrderMixin(local):

    '''Ordering knife mixin.'''

    def group(self):
        '''
        Group incoming things using :meth:`worker` as the :term:`key function`.

        :return: :func:`~collections.namedtuple` ``Group(keys=keys,
          groups=tuple)``

        :rtype: :obj:`knife` :term:`object`

        >>> from knife import __
        >>> # default grouping
        >>> __(1.3, 2.1).group().get()
        [Group(keys=1.3, groups=(1.3,)), Group(keys=2.1, groups=(2.1,))]
        >>> from math import floor
        >>> # use worker for key function
        >>> __(1.3, 2.1, 2.4).worker(floor).group().get()
        [Group(keys=1.0, groups=(1.3,)), Group(keys=2.0, groups=(2.1, 2.4))]

        .. seealso::

          :func:`itertools.groupby`
            function in Python standard library

          `groupBy <http://documentcloud.github.com/underscore/#groupBy>`_
            function in Underscore.js

          `groupBy <http://vti.github.com/underscore-perl/#groupBy>`_
            function in Underscore.perl

          `groupBy <http://brianhaveri.github.com/Underscore.php/#groupBy>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._group()

    def reverse(self):
        '''
        Reverse the order of incoming things.

        :rtype: :obj:`knife` :term:`object`

        >>> __(5, 4, 3, 2, 1).reverse().get()
        [1, 2, 3, 4, 5]

        .. seealso::

          :func:`reversed`
            function in Python standard library

          `reverse <http://mirven.github.com/underscore.lua/#reverse>`_
            function in Underscore.lua
        '''
        with self._chain:
            return self._reverse()

    def shuffle(self):
        '''
        Randomly sort incoming things.

        :rtype: :obj:`knife` :term:`object`

          >>> __(5, 4, 3, 2, 1).shuffle().get() # doctest: +SKIP
          [3, 1, 5, 4, 2]

        .. seealso::

          :func:`random.shuffle`
            function in Python standard library

          `shuffle <http://documentcloud.github.com/underscore/#shuffle>`_
            function in Underscore.js

          `shuffle <http://brianhaveri.github.com/Underscore.php/#shuffle>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._shuffle()

    def sort(self):
        '''
        Reorder incoming things using :meth:`worker` as the
        :term:`key function`.

        :rtype: :obj:`knife` :term:`object`

        >>> # default sort
        >>> __(4, 6, 65, 3, 63, 2, 4).sort().get()
        [2, 3, 4, 4, 6, 63, 65]
        >>> from math import sin
        >>> # using worker as key function
        >>> __(1, 2, 3, 4, 5, 6).worker(sin).sort().get()
        [5, 4, 6, 3, 1, 2]

        .. seealso::

          :func:`sorted`
            function in Python standard library

          `sortBy <http://documentcloud.github.com/underscore/#sortBy>`_
            function in Underscore.js

          `sort <http://mirven.github.com/underscore.lua/#sort>`_
            function in Underscore.lua

          `sortBy <http://vti.github.com/underscore-perl/#sortBy>`_
            function in Underscore.perl

          `sortBy <http://brianhaveri.github.com/Underscore.php/#sortBy>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._sort()


class RepeatMixin(local):

    '''Repeating knife mixin.'''

    def combinate(self, n):
        '''
        Discover `combinations <https://en.wikipedia.org/wiki/Combination>`_
        for every `n` incoming things.

        :argument int n: number of incoming things to derive combinations from

        :rtype: :obj:`knife` :term:`object`

        >>> from knife import __
        >>> __(40, 50, 60).combinate(2).get()
        [(40, 50), (40, 60), (50, 60)]

        .. seealso::

          :func:`itertools.combinations`
            function in Python standard library
        '''
        with self._chain:
            return self._combinate(n)

    def copy(self):
        '''
        Duplicate each incoming thing.

        :rtype: :obj:`knife` :term:`object`

        >>> __([[1, [2, 3]], [4, [5, 6]]]).copy().get()
        [[1, [2, 3]], [4, [5, 6]]]

        .. seealso::

          :func:`copy.deepcopy`
            function in Python standard library
        '''
        with self._chain:
            return self._copy()

    def permutate(self, n):
        '''
        Discover `permutations <https://en.wikipedia.org/wiki/Permutation>`_
        for every `n` incoming things.

        :argument int n: number of incoming things to derive permutations from

        :rtype: :obj:`knife` :term:`object`

        >>> __(40, 50, 60).permutate(2).get()
        [(40, 50), (40, 60), (50, 40), (50, 60), (60, 40), (60, 50)]

        .. seealso::

          :func:`itertools.permutations`
            function in Python standard library
        '''
        with self._chain:
            return self._permute(n)

    def repeat(self, n=None, call=False):
        '''
        Repeat incoming things `n` times or invoke :meth:`worker` `n` times.

        :keyword int n: number of times to repeat

        :keyword bool call: repeat results of invoking :meth:`worker`

        :rtype: :obj:`knife` :term:`object`

        >>> # repeat iterable
        >>> __(40, 50, 60).repeat(3).get()
        [(40, 50, 60), (40, 50, 60), (40, 50, 60)]
        >>> def test(*args):
        ...    return list(args)
        >>> # with worker
        >>> __(40, 50, 60).worker(test).repeat(n=3, call=True).get()
        [[40, 50, 60], [40, 50, 60], [40, 50, 60]]

        .. seealso::

          :func:`itertools.repeat`
            function in Python standard library

          `times <http://documentcloud.github.com/underscore/#times>`_
            function in Underscore.js

          `times <http://vti.github.com/underscore-perl/#times>`_
            function in Underscore.perl

          `times <http://brianhaveri.github.com/Underscore.php/#times>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._repeat(n, call)


class MapMixin(local):

    '''Mapping knife mixin.'''

    def argmap(self, merge=False):
        '''
        Feed each incoming thing to :meth:`worker` as :term:`positional
        argument`\s.

        :keyword bool merge: merge global positional :meth:`params` with
          positional arguments derived from incoming things when passed to
          :meth:`worker`

        :rtype: :obj:`knife` :term:`object`

        >>> from knife import __
        >>> # default behavior
        >>> test = __((1, 2), (2, 3), (3, 4))
        >>> test.worker(lambda x, y: x * y).argmap().get()
        [2, 6, 12]
        >>> # merge global positional arguments with iterable arguments
        >>> test.original().worker(
        ...   lambda x, y, z, a, b: x * y * z * a * b
        ... ).params(7, 8, 9).argmap(merge=True).get()
        [1008, 3024, 6048]

        .. seealso::

          :func:`itertools.starmap`
            function in Python standard library
        '''
        with self._chain:
            return self._argmap(merge)

    def invoke(self, name):
        '''
        Feed global :term:`positional argument`\s and :term:`keyword
        argument`\s to each incoming thing's `name` :term:`method`.

        .. note::

          The original thing is returned if the return value of :term:`method`
          `name` is :data:`None`.

        :argument str name: method name

        :rtype: :obj:`knife` :term:`object`

        >>> # invoke list.index()
        >>> __([5, 1, 7], [3, 2, 1]).params(1).invoke('index').get()
        [1, 2]
        >>> # invoke list.sort() but return sorted list instead of None
        >>> __([5, 1, 7], [3, 2, 1]).invoke('sort').get()
        [[1, 5, 7], [1, 2, 3]]

        .. seealso::

          `invoke <http://documentcloud.github.com/underscore/#invoke>`_
            function in Underscore.js

          `invoke <http://mirven.github.com/underscore.lua/#invoke>`_
            function in Underscore.lua

          `invoke <http://vti.github.com/underscore-perl/#invoke>`_
            function in Underscore.perl

          `invoke <http://brianhaveri.github.com/Underscore.php/#invoke>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._invoke(name)

    def kwargmap(self, merge=False):
        '''
        Feed each incoming thing as a :func:`tuple` of
        :term:`positional argument`\s and :term:`keyword argument`\s to
        :meth:`worker`.

        :keyword bool merge: merge global positional or keyword :meth:`params`
          with positional and keyword arguments derived from incoming things
          into a single :func:`tuple` of wildcard positional and keyword
          arguments like ``(*iterable_args + global_args, **global_kwargs +
          iterable_kwargs)`` when passed to :meth:`worker`

        :rtype: :obj:`knife` :term:`object`

        >>> # default behavior
        >>> test = __(
        ...  ((1, 2), {'a': 2}), ((2, 3), {'a': 2}), ((3, 4), {'a': 2})
        ... )
        >>> def tester(*args, **kw):
        ...    return sum(args) * sum(kw.values())
        >>> test.worker(tester).kwargmap().get()
        [6, 10, 14]
        >>> # merging global and iterable derived positional and keyword args
        >>> test.original().worker(tester).params(
        ...   1, 2, 3, b=5, w=10, y=13
        ... ).kwargmap(merge=True).get()
        [270, 330, 390]
        '''
        with self._chain:
            return self._kwargmap(merge)

    def map(self):
        '''
        Feed each incoming thing to :meth:`worker`.

        :rtype: :obj:`knife` :term:`object`

        >>> __(1, 2, 3).worker(lambda x: x * 3).map().get()
        [3, 6, 9]

        .. seealso::

          :func:`itertools.imap`
            function in Python standard library (replaces :func:`map` in
            Python 3)

          `map <http://documentcloud.github.com/underscore/#map>`_
            function in Underscore.js

          `map <http://mirven.github.com/underscore.lua/#map>`_
            function in Underscore.lua

          `map <http://vti.github.com/underscore-perl/#map>`_
            function in Underscore.perl

          `map <http://brianhaveri.github.com/Underscore.php/#map>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._map()

    def mapping(self, keys=False, values=False):
        '''
        Run :meth:`worker` on incoming :term:`mapping` things.

        :keyword bool keys: collect mapping keys only

        :keyword bool values: collect mapping values only

        :rtype: :obj:`knife` :term:`object`

        >>> # filter items
        >>> __(dict([(1, 2), (2, 3), (3, 4)]), dict([(1, 2), (2, 3), (3, 4)])
        ... ).worker(lambda x, y: x * y).mapping().get()
        [2, 6, 12, 2, 6, 12]
        >>> # mapping keys only
        >>> __(dict([(1, 2), (2, 3), (3, 4)]), dict([(1, 2), (2, 3), (3, 4)])
        ... ).mapping(keys=True).get()
        [1, 2, 3, 1, 2, 3]
        >>> # mapping values only
        >>> __(dict([(1, 2), (2, 3), (3, 4)]), dict([(1, 2), (2, 3), (3, 4)])
        ... ).mapping(values=True).get()
        [2, 3, 4, 2, 3, 4]

        .. seealso::

          `keys <http://documentcloud.github.com/underscore/#keys>`_
            function in Underscore.js

          `values <http://documentcloud.github.com/underscore/#values>`_
            function in Underscore.js

          `keys <http://mirven.github.com/underscore.lua/#keys>`_
            function in Underscore.lua

          `values <http://mirven.github.com/underscore.lua/#values>`_
            function in Underscore.lua

          `keys <http://vti.github.com/underscore-perl/#keys>`_
            function in Underscore.perl

          `values <http://vti.github.com/underscore-perl/#values>`_
            function in Underscore.perl

          `keys <http://brianhaveri.github.com/Underscore.php/#keys>`_
            function in Underscore.php

          `values <http://brianhaveri.github.com/Underscore.php/#values>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._mapping(keys, values)


class FilterMixin(local):

    '''Filtering knife mixin.'''

    def attrs(self, *names):
        '''
        Collect :term:`attribute` values from incoming things that match an
        **attribute name** found in `names`.

        :argument str names: attribute names

        :rtype: :obj:`knife` :term:`object`

        >>> from knife import __
        >>> from stuf import stuf
        >>> stooge = [
        ...    stuf(name='moe', age=40),
        ...    stuf(name='larry', age=50),
        ...    stuf(name='curly', age=60),
        ... ]
        >>> __(*stooge).attrs('name').get()
        ['moe', 'larry', 'curly']
        >>> # multiple attribute names
        >>> __(*stooge).attrs('name', 'age').get()
        [('moe', 40), ('larry', 50), ('curly', 60)]
        >>> # no attrs named 'place'
        >>> __(*stooge).attrs('place').get()
        []

        .. seealso::

          :func:`operator.attrgetter`
            function in Python standard library

          `pick <http://documentcloud.github.com/underscore/#pick>`_
            function in Underscore.js
        '''
        with self._chain:
            return self._attrs(names)

    def duality(self):
        '''
        Divide incoming things into two :term:`iterable`\s, the first
        everything :meth:`worker` is :data:`True` for and the second
        everything :meth:`worker` is :data:`False` for.

        :rtype: :obj:`knife` :term:`object`

        >>> test = __(1, 2, 3, 4, 5, 6).worker(lambda x: x % 2 == 0)
        >>> divide = test.duality().get()
        >>> divide.true
        (2, 4, 6)
        >>> divide.false
        (1, 3, 5)
        '''
        with self._chain:
            return self._duality()

    def filter(self, invert=False):
        '''
        Collect incoming things matched by :meth:`worker`.

        :keyword bool invert: collect incoming things :meth:`worker` is
          :data:`False` rather than :data:`True` for

        :rtype: :obj:`knife` :term:`object`

        >>> # filter for true values
        >>> test = __(1, 2, 3, 4, 5, 6).worker(lambda x: x % 2 == 0)
        >>> test.filter().get()
        [2, 4, 6]
        >>> # filter for false values
        >>> test.original().worker(
        ...   lambda x: x % 2 == 0
        ... ).filter(invert=True).get()
        [1, 3, 5]

        .. seealso::

          :func:`itertools.ifilter`
            function in Python standard library (replaces :func:`filter` in
            Python 3)

          :func:`itertools.ifilterfalse`
            function in Python standard library (:func:`filterfalse` in
            Python 3)

          `filter <http://documentcloud.github.com/underscore/#filter>`_
            function in Underscore.js

          `reject <http://documentcloud.github.com/underscore/#reject>`_
            function in Underscore.js

          `filter <http://mirven.github.com/underscore.lua/#filter>`_
            function in Underscore.lua

          `reject <http://mirven.github.com/underscore.lua/#reject>`_
            function in Underscore.lua

          `filter <http://vti.github.com/underscore-perl/#filter>`_
            function in Underscore.perl

          `reject <http://vti.github.com/underscore-perl/#reject>`_
            function in Underscore.perl

          `filter <http://brianhaveri.github.com/Underscore.php/#filter>`_
            function in Underscore.php

          `reject <http://brianhaveri.github.com/Underscore.php/#reject>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._filter(invert)

    def items(self, *keys):
        '''
        Collect values from incoming things (usually a :term:`sequence` or
        :term:`mapping`) that match a **key** found in `keys`.

        :argument str keys: keys or indices

        :rtype: :obj:`knife` :term:`object`

        >>> stooge = [
        ...    dict(name='moe', age=40),
        ...    dict(name='larry', age=50),
        ...    dict(name='curly', age=60)
        ... ]
        >>> # get items from mappings like dictionaries, etc...
        >>> __(*stooge).items('name').get()
        ['moe', 'larry', 'curly']
        >>> __(*stooge).items('name', 'age').get()
        [('moe', 40), ('larry', 50), ('curly', 60)]
        >>> # get items from sequences like lists, tuples, etc...
        >>> stooge = [['moe', 40], ['larry', 50], ['curly', 60]]
        >>> __(*stooge).items(0).get()
        ['moe', 'larry', 'curly']
        >>> __(*stooge).items(1).get()
        [40, 50, 60]
        >>> __(*stooge).items('place').get()
        []

       .. seealso::

          :func:`operator.itemgetter`
            function in Python standard library

          `pick <http://documentcloud.github.com/underscore/#pick>`_
            function in Underscore.js
        '''
        with self._chain:
            return self._items(keys)

    def traverse(self, invert=False):
        '''
        Collect values from deeply :term:`nested scope`\s from incoming things
        matched by :meth:`worker`.

        :keyword bool invert: collect incoming things that :meth:`worker` is
          :data:`False` rather than :data:`True` for

        :returns: :term:`sequence` of `ChainMaps <http://docs.python.org/dev/
          library/collections.html#collections.ChainMap>`_ containing
          :class:`collections.OrderedDict`

        :rtype: :obj:`knife` :term:`object`

        >>> class stooge:
        ...    name = 'moe'
        ...    age = 40
        >>> class stooge2:
        ...    name = 'larry'
        ...    age = 50
        >>> class stooge3:
        ...    name = 'curly'
        ...    age = 60
        ...    class stooge4(object):
        ...        name = 'beastly'
        ...        age = 969
        >>> def test(x):
        ...    if x[0] == 'name':
        ...        return True
        ...    elif x[0].startswith('__'):
        ...        return True
        ...    return False
        >>> # using worker while filtering for False values
        >>> __(stooge, stooge2, stooge3).worker(test).traverse(
        ...   invert=True
        ... ).get() # doctest: +NORMALIZE_WHITESPACE
        [ChainMap(OrderedDict([('classname', 'stooge'), ('age', 40)])),
        ChainMap(OrderedDict([('classname', 'stooge2'), ('age', 50)])),
        ChainMap(OrderedDict([('classname', 'stooge3'), ('age', 60)]),
        OrderedDict([('age', 969), ('classname', 'stooge4')]))]
        '''
        with self._chain:
            return self._traverse(invert)


class ReduceMixin(local):

    '''Reducing knife mixin.'''

    def flatten(self):
        '''
        Reduce nested incoming things to flattened incoming things.

        :rtype: :obj:`knife` :term:`object`

        >>> from knife import __
        >>> __([[1, [2], [3, [[4]]]], 'here']).flatten().get()
        [1, 2, 3, 4, 'here']

        .. seealso::

          `flatten <http://documentcloud.github.com/underscore/#flatten>`_
            function in Underscore.js

          `flatten <http://mirven.github.com/underscore.lua/#flatten>`_
            function in Underscore.lua

          `flatten <http://vti.github.com/underscore-perl/#flatten>`_
            function in Underscore.perl

          `flatten <http://brianhaveri.github.com/Underscore.php/#flatten>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._flatten()

    def merge(self):
        '''
        Reduce multiple :term:`iterable` incoming things into one iterable
        incoming thing.

        :rtype: :obj:`knife` :term:`object`

        >>> __(['moe', 'larry'], [30, 40], [True, False]).merge().get()
        ['moe', 'larry', 30, 40, True, False]

        .. seealso::

          :meth:`itertools.chain.from_iterable`
            function in Python standard library
        '''
        with self._chain:
            return self._merge()

    def reduce(self, initial=None, reverse=False):
        '''
        Reduce :term:`iterable` incoming things down to one incoming thing
        using :meth:`worker`.

        :keyword initial: starting value

        :keyword bool reverse: reduce from `the right side <http://www.zvon.
          org/other/haskell/Outputprelude/foldr_f.html>`_ of incoming things

        :rtype: :obj:`knife` :term:`object`

        >>> # reduce from left side
        >>> __(1, 2, 3).worker(lambda x, y: x + y).reduce().get()
        6
        >>> # reduce from left side with initial value
        >>> __(1, 2, 3).worker(lambda x, y: x + y).reduce(initial=1).get()
        7
        >>> # reduce from right side
        >>> test = __([0, 1], [2, 3], [4, 5]).worker(lambda x, y: x + y)
        >>> test.reduce(reverse=True).get()
        [4, 5, 2, 3, 0, 1]
        >>> # reduce from right side with initial value
        >>> test.original().worker(
        ... lambda x, y: x + y
        ... ).reduce([0, 0], True).get()
        [4, 5, 2, 3, 0, 1, 0, 0]

        .. seealso::

          :func:`functools.reduce`
            function in Python standard library

          `reduce <http://documentcloud.github.com/underscore/#reduce>`_
            function in Underscore.js

          `reduceRight <http://documentcloud.github.com/underscore/#reduceRight>`_
            function in Underscore.js

          `reduce <http://mirven.github.com/underscore.lua/#reduce>`_
            function in Underscore.lua

          `reduce <http://vti.github.com/underscore-perl/#reduce>`_
            function in Underscore.perl

          `reduceRight <http://vti.github.com/underscore-perl/#reduceRight>`_
            function in Underscore.perl

          `reduce <http://brianhaveri.github.com/Underscore.php/#reduce>`_
            function in Underscore.php

          `reduceRight <http://brianhaveri.github.com/Underscore.php/#reduceRight>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._reduce(initial, reverse)

    def zip(self):
        '''
        Convert multiple :term:`iterable` incoming things to a series of
        :func:`tuple`\s composed of things found at the same index position
        within the original iterables.

        :rtype: :obj:`knife` :term:`object`

        >>> test = __(['moe', 'larry'], [30, 40], [True, False])
        >>> test.zip().get()
        [('moe', 30, True), ('larry', 40, False)]

        .. seealso::

          :func:`itertools.izip`
            function in Python standard library (replaces :func:`zip` in
            Python 3)

          `zip <http://documentcloud.github.com/underscore/#zip>`_
            function in Underscore.js

          `zip <http://vti.github.com/underscore-perl/#zip>`_
            function in Underscore.perl

          `zip <http://brianhaveri.github.com/Underscore.php/#zip>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._zip()


class SliceMixin(local):

    '''Slicing knife mixin.'''

    def at(self, n, default=None):
        '''
        :term:`Slice` off incoming thing found at index `n`.

        :argument int n: index of some incoming thing

        :keyword default: default returned if nothing is found at `n`

        :rtype: :obj:`knife` :term:`object`

        >>> from knife import __
        >>> # default behavior
        >>> __(5, 4, 3, 2, 1).at(2).get()
        3
        >>> # return default value if nothing found at index
        >>> __(5, 4, 3, 2, 1).at(10, 11).get()
        11

        .. seealso::

          `"nth" <http://docs.python.org/library/itertools.html#recipes>`_
            Itertools Recipes
        '''
        with self._chain:
            return self._at(n, default)

    def choice(self):
        '''
        Randomly :term:`slice` off **one** incoming thing.

        :rtype: :obj:`knife` :term:`object`

        >>> __(1, 2, 3, 4, 5, 6).choice().get() # doctest: +SKIP
        3

        .. seealso::

          :func:`random.choice`
            function in Python standard library
        '''
        with self._chain:
            return self._choice()

    def dice(self, n, fill=None):
        '''
        :term:`Slice` one :term:`iterable` incoming thing into `n` iterable
        incoming things.

        :argument int n: number of incoming things per slice

        :keyword fill: value to pad out incomplete iterables

        :rtype: :obj:`knife` :term:`object`

        >>> __('moe', 'larry', 'curly', 30, 40, 50, True).dice(2, 'x').get()
        [('moe', 'larry'), ('curly', 30), (40, 50), (True, 'x')]

        .. seealso::

          `"grouper" <http://docs.python.org/library/itertools.html#recipes>`_
            Itertools Recipes
        '''
        with self._chain:
            return self._dice(n, fill)

    def first(self, n=0):
        '''
        :term:`Slice`  off `n` things from the **starting** end of incoming
        things or just the **first** incoming thing.

        :keyword int n: number of incoming things

        :rtype: :obj:`knife` :term:`object`

        >>> # default behavior
        >>> __(5, 4, 3, 2, 1).first().get()
        5
        >>> # first things from index 0 to 2
        >>> __(5, 4, 3, 2, 1).first(2).get()
        [5, 4]

        .. seealso::

          `first <http://documentcloud.github.com/underscore/#first>`_
            function in Underscore.js

          `first <http://mirven.github.com/underscore.lua/#first>`_
            function in Underscore.lua

          `first <http://vti.github.com/underscore-perl/#first>`_
            function in Underscore.perl

          `first <http://brianhaveri.github.com/Underscore.php/#first>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._first(n)

    def initial(self):
        '''
        :term:`Slice` off every incoming thing except the **last** incoming
        thing.

        :rtype: :obj:`knife` :term:`object`

        >>> __(5, 4, 3, 2, 1).initial().get()
        [5, 4, 3, 2]

        .. seealso::

          `initial <http://documentcloud.github.com/underscore/#initial>`_
            function in Underscore.js

          `initial <http://mirven.github.com/underscore.lua/#initial>`_
            function in Underscore.lua

          `initial <http://vti.github.com/underscore-perl/#initial>`_
            function in Underscore.perl

          `initial <http://brianhaveri.github.com/Underscore.php/#initial>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._initial()

    def last(self, n=0):
        '''
        :term:`Slice` off `n` things from the **tail** end of incoming things
        or just the **last** incoming thing.

        :keyword int n: number of incoming things

        :rtype: :obj:`knife` :term:`object`

        >>> # default behavior
        >>> __(5, 4, 3, 2, 1).last().get()
        1
        >>> # fetch last two things
        >>> __(5, 4, 3, 2, 1).last(2).get()
        [2, 1]

        .. seealso::

          `last <http://documentcloud.github.com/underscore/#last>`_
            function in Underscore.js

          `last <http://mirven.github.com/underscore.lua/#last>`_
            function in Underscore.lua

          `last <http://vti.github.com/underscore-perl/#last>`_
            function in Underscore.perl

          `last <http://brianhaveri.github.com/Underscore.php/#last>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._last(n)

    def rest(self):
        '''
        :term:`Slice` off every incoming thing except the **first** incoming
        thing.

        :rtype: :mod:`knife` :term:`object`

        >>> __(5, 4, 3, 2, 1).rest().get()
        [4, 3, 2, 1]

        .. seealso::

          `rest <http://documentcloud.github.com/underscore/#rest>`_
            function in Underscore.js

          `rest <http://mirven.github.com/underscore.lua/#rest>`_
            function in Underscore.lua

          `rest <http://vti.github.com/underscore-perl/#rest>`_
            function in Underscore.perl

          `rest <http://brianhaveri.github.com/Underscore.php/#rest>`_
            function in Underscore.php
        '''
        with self._chain:
            return self._rest()

    def sample(self, n):
        '''
        Randomly :term:`slice` off `n` incoming things.

        :argument int n: sample size

        :rtype: :mod:`knife` :term:`object`

        >>> __(1, 2, 3, 4, 5, 6).sample(3).get() # doctest: +SKIP
        [2, 4, 5]

        .. seealso::

          :func:`random.sample`
            function in Python standard library
        '''
        with self._chain:
            return self._sample(n)

    def slice(self, start, stop=False, step=False):
        '''
        Take :term:`slice` out of incoming things.

        :argument int start: starting index of slice

        :keyword int stop: stopping index of slice

        :keyword int step: size of step in slice

        :rtype: :mod:`knife` :term:`object`

        >>> # slice from index 0 to 3
        >>> __(5, 4, 3, 2, 1).slice(2).get()
        [5, 4]
        >>> # slice from index 2 to 4
        >>> __(5, 4, 3, 2, 1).slice(2, 4).get()
        [3, 2]
        >>> # slice from index 2 to 4 with 2 steps
        >>> __(5, 4, 3, 2, 1).slice(2, 4, 2).get()
        3

        .. seealso::

          :func:`itertools.islice`
            function in Python standard library

          `slice <http://mirven.github.com/underscore.lua/#slice>`_
            function in Underscore.lua
        '''
        with self._chain:
            return self._slice(start, stop, step)



# -*- coding: utf-8 -*-
'''active knife tests'''

from stuf.six import unittest

from tests.mixins import (
    Mixin, MapMixin, RepeatMixin, ReduceMixin, SliceMixin, FilterMixin,
    OrderMixin, MathMixin, CmpMixin)


class TestActive(
    unittest.TestCase, Mixin, FilterMixin, MapMixin, ReduceMixin, OrderMixin,
    SliceMixin, RepeatMixin, MathMixin, CmpMixin,
):

    def setUp(self):
        from knife import activeknife
        self.mclass = activeknife
        self.pipe = activeknife


class TestCompare(unittest.TestCase, Mixin, CmpMixin):

    def setUp(self):
        from knife.active import cmpknife
        self.mclass = cmpknife


class TestFilter(unittest.TestCase, Mixin, FilterMixin):

    def setUp(self):
        self.maxDiff = None
        from knife.active import filterknife
        self.mclass = filterknife


class TestMap(unittest.TestCase, Mixin, MapMixin):

    def setUp(self):
        from knife.active import mapknife
        self.mclass = mapknife


class TestMath(unittest.TestCase, Mixin, MathMixin):

    def setUp(self):
        from knife.active import mathknife, reduceknife
        self.mclass = mathknife
        self.pipe = reduceknife


class TestOrder(unittest.TestCase, Mixin, OrderMixin):

    def setUp(self):
        from knife.active import orderknife
        self.mclass = orderknife


class TestReduce(unittest.TestCase, Mixin, ReduceMixin):

    def setUp(self):
        from knife.active import reduceknife
        self.mclass = reduceknife


class TestRepeat(unittest.TestCase, Mixin, RepeatMixin):

    def setUp(self):
        from knife.active import repeatknife
        self.mclass = repeatknife


class TestSlice(unittest.TestCase, Mixin, SliceMixin):

    def setUp(self):
        from knife.active import sliceknife
        self.mclass = sliceknife


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
'''base knife mixins'''

from threading import local

from stuf.six import tounicode, tobytes


class KnifeMixin(local):

    '''Base knife mixin.'''

    def apply(self, worker, *args, **kw):
        '''
        Assign :func:`callable` used to work on incoming things plus any
        :term:`positional argument`\s and :term:`keyword argument`\s
        it will use.

        .. note::

          Global :term:`positional argument`\s and :term:`keyword argument`\s
          assigned with :meth:`params` are reset whenever :func:`apply` is
          called.

        :argument worker: a :func:`callable`

        :rtype: :mod:`knife` :term:`object`
        '''
        # assign worker
        self._worker = worker
        # positional params
        self._args = args
        # keyword arguemnts
        self._kw = kw
        return self

    def worker(self, worker):
        '''
        Assign :func:`callable` used to work on incoming things.

        .. note::

          Global :term:`positional argument`\s and :term:`keyword argument`\s
          assigned with :meth:`params` are reset whenever a new `worker` is
          assigned.

        :argument worker: a :func:`callable`

        :rtype: :mod:`knife` :term:`object`
        '''
        # reset stored position params
        self._args = ()
        # reset stored keyword params
        self._kw = {}
        # assign worker
        self._worker = worker
        return self

    def params(self, *args, **kw):
        '''
        Assign :term:`positional argument`\s and :term:`keyword argument`\s
        to be used globally.

        :rtype: :mod:`knife` :term:`object`
        '''
        # positional params
        self._args = args
        # keyword arguemnts
        self._kw = kw
        return self

    def pattern(self, pattern, type='parse', flags=0):
        '''
        Compile search `pattern` for use as :meth:`worker`.

        .. note::

          Global :term:`positional argument`\s and :term:`keyword argument`\s
          assigned with :meth:`params` are reset whenever a new `pattern` is
          compiled.

        :argument str pattern: search pattern

        :keyword str type: engine to compile `pattern` with. Valid options
          are `'parse' <http://pypi.python.org/pypi/parse/>`_, `'re'
          <http://docs.python.org/library/re.html>`_, or `'glob' <http://docs.
          python.org/library/fnmatch.html>`_
        :keyword int flags: regular expression `flags <http://docs.python.org/
          library/re.html#re.DEBUG>`_

        :rtype: :mod:`knife` :term:`object`

        >>> # using parse expression
        >>> test = __('first test', 'second test', 'third test')
        >>> test.pattern('first {}').filter().get()
        'first test'
        >>> # using glob pattern
        >>> test.original().pattern('second*', type='glob').filter().get()
        'second test'
        >>> # using regular expression
        >>> test.original().pattern('third .', type='regex').filter().get()
        'third test'
        '''
        # reset stored position params
        self._args = ()
        # reset stored keyword params
        self._kw = {}
        self._worker = self._pattern(pattern, type, flags)
        return self

    def prepend(self, *things):
        '''
        Insert `things` **before** other incoming things.

        :argument things: incoming things

        :rtype: :mod:`knife` :term:`object`

        >>> __(3, 4, 5).prepend(1, 2, 3, 4, 5, 6).peek()
        [1, 2, 3, 4, 5, 6, 3, 4, 5]
        '''
        return self._prependit(things)

    def append(self, *things):
        '''
        Insert `things` **after** other incoming things.

        :argument things: incoming things

        :rtype: :mod:`knife` :term:`object`

        >>> from knife import __
        >>> __(3, 4, 5).append(1, 2, 3, 4, 5, 6).peek()
        [3, 4, 5, 1, 2, 3, 4, 5, 6]
        '''
        return self._appendit(things)

    def pipe(self, knife):
        '''
        Pipe incoming things from some other :mod:`knife` :term:`object`
        through this :mod:`knife` :term:`object`.

        :argument knife: another :mod:`knife` :term:`object`

        :rtype: :mod:`knife` :term:`object`
        '''
        with self._chain:
            return self._pipeit(knife)

    def back(self):
        '''
        Switch back to :mod:`knife` :term:`object` that piped its incoming
        things through this :mod:`knife` :term:`object`.

        :rtype: :mod:`knife` :term:`object`
        '''
        with self._chain:
            return self._unpipeit()

    def __len__(self):
        '''Number of incoming things.'''
        return self._len()

    def __repr__(self):
        '''String representation.'''
        return self._repr()


class OutMixin(KnifeMixin):

    '''Output mixin.'''

    def __iter__(self):
        '''Iterate over outgoing things.'''
        return self._iterate()

    def get(self):
        '''
        Return outgoing things wrapped with :meth:`wrap`.

        .. note::

          With only one outgoing thing, only that one outgoing thing is
          returned. With multiple outgoing things, they are returned wrapped
          with the wrapper assigned with :meth:`wrap` (default wrapper is
          :func:`list`).
        '''
        return self._get()

    def peek(self):
        '''
        Preview current incoming things wrapped with :meth:`wrap`.

        .. note::

          With only one outgoing thing, only that one thing is returned. With
          multiple outgoing things, everything is returned wrapped with the
          wrapper assigned with :meth:`wrap` (default wrapper is :func:`list`).
        '''
        return self._peek()

    def wrap(self, wrapper):
        '''
        Assign :term:`object`, :term:`type`, or :keyword:`class` used to wrap
        outgoing things.

        .. note::

          A :mod:`knife` :term:`object` resets back to its default wrapper
          (:func:`list`) after :meth:`get` or :meth:`peek` is called.

        :argument wrapper: an :term:`object`, :func:`type`, or :keyword:`class`

        :rtype: :mod:`knife` :term:`object`

        >>> __(1, 2, 3, 4, 5, 6).wrap(tuple).peek()
        (1, 2, 3, 4, 5, 6)
        '''
        self._wrapper = wrapper
        return self

    def oneach(self):
        '''
        Toggle whether each outgoing thing should be individually wrapped with
        the wrapper assigned with :meth:`wrap` (default wrapper is :func:`list`
        ) or whether all outgoing things should be wrapped all at once.

        .. note::

          :mod:`knife` :term:`object` default behavior is to wrap all outgoing
          things all at once. :mod:`knife` :term:`object`\s reset back to this
          behavior **after** :meth:`get` or :meth:`peek` is called.

        :rtype: :mod:`knife` :term:`object`
        '''
        self._each = not self._each
        return self

    def ascii(self, errors='strict'):
        '''
        :meth:`~str.encode` outgoing things as `bytes <http://docs.python
        .org/py3k/library/functions.html#bytes>`_ with the ``'latin-1'`` codec.

        :keyword str errors: `error handling <http://docs.python.org/library/
          codecs.html#codec-base-classes>`_ for encoding

        :rtype: :mod:`knife` :term:`object`

        >>> from stuf.six import u, b
        >>> test = __([1], True, r't', b('i'), u('g'), None, (1,))
        >>> test.ascii().oneach().peek()
        ['[1]', 'True', 't', 'i', 'g', 'None', '(1,)']
        '''
        self._wrapper = lambda x: tobytes(x, 'latin_1', errors)
        return self

    def bytes(self, encoding='utf_8', errors='strict'):
        '''
        :meth:`~str.encode` outgoing things as `bytes <http://docs.python
        .org/py3k/library/functions.html#bytes>`_.

        :keyword str encoding: character `encoding <http://docs.python.org/
          library/codecs.html#standard-encodings>`_

        :keyword str errors: `error handling <http://docs.python.org/library/
          codecs.html#codec-base-classes>`_ for encoding

        :rtype: :mod:`knife` :term:`object`

        >>> test = __([1], True, r't', b('i'), u('g'), None, (1,))
        >>> test.bytes().oneach().peek()
        ['[1]', 'True', 't', 'i', 'g', 'None', '(1,)']
        '''
        self._wrapper = lambda x: tobytes(x, encoding, errors)
        return self

    def unicode(self, encoding='utf_8', errors='strict'):
        '''
        :func:`unicode` (:func:`str` under Python 3) :meth:`~str.decode`
        outgoing things.

        :keyword str encoding: Unicode `encoding <http://docs.python.org/
          library/codecs.html#standard-encodings>`_

        :keyword str errors: `error handling <http://docs.python.org/library/
          codecs.html#codec-base-classes>`_ for decoding

        :rtype: :mod:`knife` :term:`object`

        >>> test = __([1], True, r't', b('i'), u('g'), None, (1,))
        >>> test.unicode().oneach().peek()
        [u'[1]', u'True', u't', u'i', u'g', u'None', u'(1,)']
        '''
        self._wrapper = lambda x: tounicode(x, encoding, errors)
        return self

    def undo(self, snapshot=0):
        '''
        Revert incoming things to a previous snapshot.

        .. note::

          A snapshot of current incoming things is taken when a :mod:`knife`
          :term:`method` is called but before the main body of the method
          executes.

        :keyword int snapshot: number of steps ago ``1``, ``2``, ``3``, etc.

        :rtype: :mod:`knife` :term:`object`

        >>> undone = __(1, 2, 3).prepend(1, 2, 3, 4, 5, 6)
        >>> undone.peek()
        [1, 2, 3, 4, 5, 6, 1, 2, 3]
        >>> # undo back one step
        >>> undone.append(1).undo().peek()
        [1, 2, 3, 4, 5, 6, 1, 2, 3]
        >>> # undo back one step
        >>> undone.append(1).append(2).undo().peek()
        [1, 2, 3, 4, 5, 6, 1, 2, 3, 1]
        >>> # undo back 2 steps
        >>> undone.append(1).append(2).undo(2).peek()
        [1, 2, 3, 4, 5, 6, 1, 2, 3, 1]
        '''
        return self._undo(snapshot)

    def snapshot(self):
        '''
        Take a baseline snapshot of current incoming things.

        :rtype: :mod:`knife` :term:`object`
        '''
        return self._snapshot()

    def baseline(self):
        '''
        Restore incoming things back to the baseline :meth:`snapshot`.

        :rtype: :mod:`knife` :term:`object`

        >>> from knife import __
        >>> undone = __(1, 2, 3).prepend(1, 2, 3, 4, 5, 6)
        >>> undone.peek()
        [1, 2, 3, 4, 5, 6, 1, 2, 3]
        >>> undone.snapshot().append(1).append(2).peek()
        [1, 2, 3, 4, 5, 6, 1, 2, 3, 1, 2]
        >>> undone.baseline().peek()
        [1, 2, 3, 4, 5, 6, 1, 2, 3]
        '''
        return self._rollback()

    def original(self):
        '''
        Restore incoming things back to the original snapshot.

        .. note::

          THe original snapshot of incoming things is taken following the first
          :mod:`knife` :term:`method` call but before the second :mod:`knife`
          :term:`method` call (if there is a second :term:`method` call)

        :rtype: :mod:`knife` :term:`object`

        >>> undone = __(1, 2, 3).prepend(1, 2, 3, 4, 5, 6)
        >>> undone.peek()
        [1, 2, 3, 4, 5, 6, 1, 2, 3]
        >>> undone.original().peek()
        [1, 2, 3]
        '''
        return self._revert()

    def clear(self):
        '''
        Clear everything.

        :rtype: :mod:`knife` :term:`object`
        '''
        return self._clear()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''setup for knife'''

from os import getcwd
from os.path import join

from setuptools import setup, find_packages


def getversion(fname):
    '''
    Get the __version__ without importing.
    '''
    for line in open(fname):
        if line.startswith('__version__'):
            return '%s.%s.%s' % eval(line[13:])

setup(
    name='knife',
    version=getversion('knife/__init__.py'),
    description='Pythonic remix of underscore.js: Things go in. Things get '
        'knifed. Things go out.',
    long_description=open(join(getcwd(), 'README.rst'), 'r').read(),
    keywords='pipeline filtering chaining iterator functional fluent chaining',
    license='BSD',
    author='L. C. Rees',
    author_email='lcrees@gmail.com',
    url='https://bitbucket.org/lcrees/knife',
    packages=find_packages(),
    test_suite='knife.tests',
    zip_safe=False,
    install_requires=list(l.strip() for l in open(
        join(getcwd(), 'reqs/requires.txt'), 'r',
    ).readlines()),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
    ],
)



# -*- coding: utf-8 -*-
'''active knives'''

from threading import local
from collections import deque
from contextlib import contextmanager

from stuf.deep import clsname
from stuf.utils import loads, optimize, moptimize


class _ActiveMixin(local):

    '''active knife mixin'''

    def __init__(self, *things, **kw):
        '''
        :argument things: incoming things

        :keyword integer snapshots: snapshots to keep (default: ``5``)
        '''
        incoming = deque()
        incoming.extend(things)
        super(_ActiveMixin, self).__init__(incoming, deque(), **kw)
        # working things
        self._work = deque()
        # holding things
        self._hold = deque()

    @property
    @contextmanager
    def _chain(self, d=moptimize):
        # take snapshot
        snapshot = d(self._in)
        # rebalance incoming with outcoming
        if self._history:
            self._in.clear()
            self._in.extend(self._out)
        # make snapshot original snapshot?
        else:
            self._original = snapshot
        # place snapshot at beginning of snapshot stack
        self._history.appendleft(snapshot)
        # move incoming things to working things
        self._work.extend(self._in)
        yield
        out = self._out
        # clear outgoing things
        out.clear()
        # extend outgoing things with holding things
        out.extend(self._hold)
        # clear working things
        self._work.clear()
        # clear holding things
        self._hold.clear()

    @property
    def _iterable(self):
        # derived from Raymond Hettinger Python Cookbook recipe # 577155
        call = self._work.popleft
        try:
            while 1:
                yield call()
        except IndexError:
            pass

    def _append(self, thing):
        # append thing after other holding things
        self._hold.append(thing)
        return self

    def _xtend(self, things):
        # place things after holding things
        self._hold.extend(things)
        return self

    def _prependit(self, things, d=moptimize):
        # take snapshot
        snapshot = d(self._in)
        # make snapshot original snapshot?
        if self._original is None:
            self._original = snapshot
        # place snapshot at beginning of snapshot stack
        self._history.appendleft(snapshot)
        # place thing before other holding things
        self._in.extendleft(reversed(things))
        return self

    def _appendit(self, things, d=moptimize):
        # take snapshot
        snapshot = d(self._in)
        # make snapshot original snapshot?
        if self._original is None:
            self._original = snapshot
        # place snapshot at beginning of snapshot stack
        self._history.appendleft(snapshot)
        # place things after other incoming things
        self._in.extend(things)
        return self

    def _pipeit(self, knife):
        knife.clear()
        knife._history.clear()
        knife._history.extend(self._history)
        knife._original = self._original
        knife._baseline = self._baseline
        knife._out.extend(self._out)
        knife._worker = self._worker
        knife._args = self._args
        knife._kw = self._kw
        knife._wrapper = self._wrapper
        knife._pipe = self
        return knife

    def _unpipeit(self):
        piped = self._pipe
        piped.clear()
        piped._history.clear()
        piped._history.extend(self._history)
        piped._original = self._original
        piped._baseline = self._baseline
        piped._out.extend(self._out)
        piped._worker = self._worker
        piped._args = self._args
        piped._kw = self._kw
        piped._wrapper = self._wrapper
        self.clear()
        return piped

    def _repr(self, clsname_=clsname, list_=list):
        # object representation
        return self._REPR.format(
            self.__module__,
            clsname_(self),
            list_(self._in),
            list_(self._work),
            list_(self._hold),
            list_(self._out),
        )

    def _len(self, len=len):
        # length of incoming things
        return len(self._in)


class _OutMixin(_ActiveMixin):

    '''active output mixin'''

    def _undo(self, snapshot=0, loads_=loads):
        # clear everything
        self.clear()
        # if specified, use a specific snapshot
        if snapshot:
            self._history.rotate(-(snapshot - 1))
        self._in.extend(loads_(self._history.popleft()))
        return self

    def _snapshot(self, d=optimize):
        # take baseline snapshot of incoming things
        self._baseline = d(self._in)
        return self

    def _rollback(self, loads_=loads):
        # clear everything
        self.clear()
        # clear snapshots
        self._clearsp()
        # revert to baseline snapshot of incoming things
        self._in.extend(loads_(self._baseline))
        return self

    def _revert(self, loads_=loads):
        # clear everything
        self.clear()
        # clear snapshots
        self._clearsp()
        # clear baseline
        self._baseline = None
        # restore original snapshot of incoming things
        self._in.extend(loads_(self._original))
        return self

    def _clear(self, list_=list):
        # clear worker
        self._worker = None
        # clear worker positional arguments
        self._args = ()
        # clear worker keyword arguments
        self._kw = {}
        # default iterable wrapper
        self._wrapper = list_
        # clear pipe
        self._pipe = None
        # clear incoming things
        self._in.clear()
        # clear working things
        self._work.clear()
        # clear holding things
        self._hold.clear()
        # clear outgoing things
        self._out.clear()
        return self

    def _iterate(self, iter_=iter):
        return iter_(self._out)

    def _peek(self, len_=len, list_=list):
        wrap, out = self._wrapper, self._in
        value = list_(wrap(i) for i in out) if self._each else wrap(out)
        self._each = False
        self._wrapper = list_
        return value[0] if len_(value) == 1 else value

    def _get(self, len_=len, list_=list):
        wrap, out = self._wrapper, self._out
        value = list_(wrap(i) for i in out) if self._each else wrap(out)
        self._each = False
        self._wrapper = list_
        return value[0] if len_(value) == 1 else value



# -*- coding: utf-8 -*-
'''Actively evaluated knives.'''

from knife.base import OutMixin
from knife.mixins import (
    RepeatMixin, MapMixin, SliceMixin, ReduceMixin, FilterMixin, MathMixin,
    CmpMixin, OrderMixin)

from knife._active import _OutMixin
from knife._base import SLOTS, _KnifeMixin
from knife._mixins import (
    _RepeatMixin, _MapMixin, _SliceMixin, _ReduceMixin, _FilterMixin,
    _MathMixin, _CmpMixin, _OrderMixin)


class activeknife(
    _OutMixin, _KnifeMixin, _CmpMixin, _FilterMixin, _MapMixin, _MathMixin,
    _OrderMixin, _ReduceMixin, _SliceMixin, _RepeatMixin,
    OutMixin, FilterMixin, MapMixin, ReduceMixin, OrderMixin, RepeatMixin,
    MathMixin, SliceMixin, CmpMixin,
):

    '''
    Actively evaluated combo knife. Provides every :mod:`knife` method.

    .. note::

      Also aliased as :class:`~knife.knife` when imported from :mod:`knife`.

    >>> from knife import knife
    '''

    __slots__ = SLOTS


class cmpknife(_OutMixin, _KnifeMixin, OutMixin, CmpMixin, _CmpMixin):

    '''
    Actively evaluated comparing knife. Provides comparison operations for
    incoming things.

    >>> from knife.active import cmpknife
    '''

    __slots__ = SLOTS


class filterknife(_OutMixin, _KnifeMixin, OutMixin, FilterMixin, _FilterMixin):

    '''
    Actively evaluated filtering knife. Provides filtering operations for
    incoming things.

    >>> from knife.active import filterknife
    '''

    __slots__ = SLOTS


class mapknife(_OutMixin, _KnifeMixin, OutMixin, MapMixin, _MapMixin):

    '''
    Actively evaluated mapping knife. Provides `mapping <http://docs.python.org
    /library/functions.html#map>`_ operations for incoming things.

    >>> from knife.active import mapknife
    '''

    __slots__ = SLOTS


class mathknife(_OutMixin, _KnifeMixin, OutMixin, MathMixin, _MathMixin):

    '''
    Actively evaluated mathing knife. Provides numeric and statistical
    operations for incoming things.

    >>> from knife.active import mathknife
    '''

    __slots__ = SLOTS


class orderknife(_OutMixin, _KnifeMixin, OutMixin, OrderMixin, _OrderMixin):

    '''
    Actively evaluated ordering knife. Provides sorting and grouping operations
    for incoming things.

    >>> from knife.active import orderknife
    '''

    __slots__ = SLOTS


class reduceknife(_OutMixin, _KnifeMixin, OutMixin, ReduceMixin, _ReduceMixin):

    '''
    Actively evaluated reducing knife. Provides `reducing <http://docs.python.
    org/library/functions.html#map>`_ operations for incoming things.

    >>> from knife.active import reduceknife
    '''

    __slots__ = SLOTS


class repeatknife(_OutMixin, _KnifeMixin, OutMixin, RepeatMixin, _RepeatMixin):

    '''
    Actively evaluated repeating knife. Provides repetition operations for
    incoming things.

    >>> from knife.active import repeatknife
    '''

    __slots__ = SLOTS


class sliceknife(_OutMixin, _KnifeMixin, OutMixin, SliceMixin, _SliceMixin):

    '''
    Actively evaluated slicing knife. Provides `slicing <http://docs.python.
    org/library/functions.html#slice>`_ operations for incoming things.

    >>> from knife.active import sliceknife
    '''

    __slots__ = SLOTS
